from dataclasses import dataclass
from typing import Optional

from omegaconf import DictConfig, OmegaConf

import autrainer
from autrainer.core.scripts.abstract_script import MockParser

from .abstract_preprocess_script import (
    AbstractPreprocessScript,
    PreprocessArgs,
)
from .utils import (
    add_hydra_args_to_sys,
    catch_cli_errors,
    check_invalid_config_path_arg,
    run_hydra_cmd,
    running_in_notebook,
)


@dataclass
class PreprocessArgs(PreprocessArgs):
    num_workers: int
    update_frequency: int


def preprocess_main(
    name: str,
    dataset: DictConfig,
    preprocess: DictConfig,
    num_workers: int,
    update_frequency: int,
):
    from pathlib import Path

    import torch
    from tqdm import tqdm

    from autrainer.core.structs import DataBatch
    from autrainer.datasets import AbstractDataset
    from autrainer.datasets.utils import AbstractFileHandler
    from autrainer.transforms import SmartCompose

    print(f" - {name}")
    if preprocess is None:
        print("No preprocessing specified. Skipping...")
        return
    # set dataset file handler as output handler
    # as dataset is already configured
    # to work with the output of the preprocessing script
    output_file_handler = autrainer.instantiate_shorthand(
        config=dataset["file_handler"],
        instance_of=AbstractFileHandler,
    )
    output_file_type = dataset["file_type"]

    # override dataset file handling to work with raw audio
    dataset["file_handler"] = preprocess["file_handler"]
    # None allows dataset to work with all audio files
    dataset["file_type"] = None
    dataset["seed"] = 0  # ignored
    dataset.pop("criterion")
    dataset.pop("transform")
    for subset in ["train", "dev", "test"]:
        dataset.pop(f"{subset}_loader_kwargs", None)
    features_path = dataset.pop("features_path", None)
    if features_path is None:
        features_path = dataset["path"]
    features_subdir = dataset["features_subdir"]

    dataset["features_subdir"] = None
    data = autrainer.instantiate(dataset, AbstractDataset)
    pipeline = SmartCompose(
        [autrainer.instantiate_shorthand(t) for t in preprocess["pipeline"]]
    )
    # use pipeline to process the features
    data.train_transform = pipeline
    data.dev_transform = pipeline
    data.test_transform = pipeline

    for d, df, n in (
        (data.train_dataset, data.df_train, "train"),
        (data.dev_dataset, data.df_dev, "dev"),
        (data.test_dataset, data.df_test, "test"),
    ):
        # TODO: dataloader underutilized
        # workers only parallelize loading
        loader = torch.utils.data.DataLoader(
            dataset=d,
            shuffle=False,
            num_workers=num_workers,
            batch_size=1,
            collate_fn=DataBatch.collate,
        )
        for instance in tqdm(
            loader,
            total=len(loader),
            desc=f"{name}-{n}",
            disable=update_frequency == 0,
        ):
            item_path = df.loc[df.index[int(instance.index)], d.index_column]
            out_path = Path(
                features_path,
                features_subdir,
                item_path,
            ).with_suffix(f".{output_file_type}")
            out_path.parent.mkdir(parents=True, exist_ok=True)
            if out_path.exists():
                continue
            output_file_handler.save(
                out_path,
                instance.features.squeeze(dim=0),
            )


class PreprocessScript(AbstractPreprocessScript):
    def __init__(self) -> None:
        super().__init__(
            "preprocess",
            "Launch a data preprocessing configuration (Hydra).",
            extended_description=(
                "For more information on Hydra's command line line flags, see:\n"
                "https://hydra.cc/docs/advanced/hydra-command-line-flags/."
            ),
            epilog="Example: autrainer preprocess -cn config.yaml",
            dataclass=PreprocessArgs,
            unknown_args=True,
        )

    def add_arguments(self) -> None:
        super().add_arguments()
        self.parser.add_argument(
            "-n",
            "--num-workers",
            type=int,
            default=0,
            metavar="N",
            required=False,
            help=(
                "Number of workers (subprocesses) to use for preprocessing. "
                "Defaults to 0."
            ),
        )
        self.parser.add_argument(
            "-u",
            "--update-frequency",
            type=int,
            default=1,
            metavar="F",
            required=False,
            help=(
                "Frequency of progress bar updates for each worker (subprocess). "
                "If 0, the progress bar will be disabled. Defaults to 1."
            ),
        )

    def main(self, args: PreprocessArgs) -> None:
        self._assert_num_workers(args.num_workers)

        import hydra

        self.num_workers = args.num_workers
        self.update_frequency = args.update_frequency
        self._override_launcher(args)
        self.datasets = {}
        self.preprocessing = {}

        @autrainer.main("config")
        def main(cfg: DictConfig) -> None:
            from hydra.errors import MissingConfigException

            if self._id_in_dict(self.datasets, cfg.dataset.id):
                return

            self.datasets[cfg.dataset.id] = OmegaConf.to_container(cfg.dataset)
            preprocessing_cfg = None
            if cfg.dataset.get("features_subdir"):
                try:
                    preprocessing_cfg = OmegaConf.to_container(
                        hydra.compose(
                            f"preprocessing/{cfg.dataset.features_subdir}"
                        )["preprocessing"]
                    )
                except MissingConfigException:
                    pass
            self.preprocessing[cfg.dataset.id] = preprocessing_cfg

        check_invalid_config_path_arg(self.parser)
        main()
        self._preprocess_datasets()
        self._clean_up()

    def _assert_num_workers(self, num_workers: int) -> None:
        if num_workers < 0:
            raise ValueError(
                f"Number of workers '{num_workers}' must be >= 0."
            )

    def _preprocess_datasets(self) -> None:
        print("Preprocessing datasets...")
        for (name, dataset), preprocess in zip(
            self.datasets.items(), self.preprocessing.values()
        ):
            preprocess_main(
                name=name,
                dataset=dataset,
                preprocess=preprocess,
                num_workers=self.num_workers,
                update_frequency=self.update_frequency,
            )


@catch_cli_errors
def preprocess(
    override_kwargs: Optional[dict] = None,
    num_workers: int = 0,
    update_frequency: int = 1,
    cfg_launcher: bool = False,
    config_name: str = "config",
    config_path: Optional[str] = None,
) -> None:
    """Launch a data preprocessing configuration.

    Args:
        override_kwargs: Additional Hydra override arguments to pass to the
            train script.
        num_workers: Number of workers (subprocess) to use for preprocessing.
            Defaults to 0.
        update_frequency: Frequency of progress bar updates.
            If 0, the progress bar will be disabled. Defaults to 1.
        cfg_launcher: Use the launcher specified in the configuration instead
            of the Hydra basic launcher. Defaults to False.
        config_name: The name of the config (usually the file name without the
            .yaml extension). Defaults to "config".
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
    """
    if running_in_notebook():
        cmd = "preprocess"
        if cfg_launcher:
            cmd += " -l"
        cmd += f" -n {num_workers} -u {update_frequency}"
        run_hydra_cmd(cmd, override_kwargs, config_name, config_path)

    else:
        add_hydra_args_to_sys(override_kwargs, config_name, config_path)
        script = PreprocessScript()
        script.parser = MockParser()
        script.main(
            PreprocessArgs(cfg_launcher, num_workers, update_frequency)
        )



from typing import List, Tuple

import torch
import torchvision

from .abstract_model import AbstractModel
from .utils import ExtractLayerEmbeddings


class TorchvisionModel(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        torchvision_name: str,
        transfer: bool = False,
        **kwargs,
    ) -> None:
        """Wrapper for torchvision models.

        Args:
            output_dim: Number of output classes.
            torchvision_name: Name of the model available in
                `torchvision.models`.
            transfer: Whether to load the model with pretrained weights.
                The "DEFAULT" weights are used if `transfer` is True.
                The final layer is replaced with a new layer with `output_dim`
                output features. Defaults to False.
            kwargs: Additional arguments to pass to the model constructor.
        """
        super().__init__(output_dim, transfer)
        self.torchvision_name = torchvision_name
        for key, value in kwargs.items():
            setattr(self, key, value)  # pragma: no cover

        if transfer:  # pragma: no cover
            self.model = torchvision.models.get_model(
                torchvision_name,
                weights="DEFAULT",
                **kwargs,
            )
            self._replace_final_layer(self.model)
        else:
            self.model = torchvision.models.get_model(
                torchvision_name,
                weights=None,
                num_classes=output_dim,
                **kwargs,
            )
        self._embedding_extractor = ExtractLayerEmbeddings(self.model)

    def _replace_final_layer(self, module: torch.nn.Module) -> None:
        """Replace the final linear layer of the model with a new linear
        layer having `self.output_dim` output features.

        Args:
            module: Model to replace the final layer.

        Raises:
            ValueError: If no linear layer is found in the model.
        """
        layers: List[Tuple[torch.nn.Module, str, torch.nn.Linear]] = []

        def collect_layers(module: torch.nn.Module):
            for name, layer in module.named_children():
                if isinstance(layer, torch.nn.Linear):
                    layers.append((module, name, layer))
                else:
                    collect_layers(layer)

        collect_layers(module)
        if layers:
            module, name, layer = layers[-1]
            in_features = layer.in_features
            new_layer = torch.nn.Linear(in_features, self.output_dim)
            setattr(module, name, new_layer)
        else:
            raise ValueError(
                f"Failed to override last linear layer for model "
                f"'{self.torchvision_name}'."
            )  # pragma: no cover

    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        return self._embedding_extractor(features)

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        return self.model(features)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List, Optional


if TYPE_CHECKING:
    from .training import ModularTaskTrainer  # pragma: no cover


CALLBACK_FUNCTIONS = [
    # training loop callbacks
    "cb_on_train_begin",
    "cb_on_train_end",
    "cb_on_iteration_begin",
    "cb_on_iteration_end",
    "cb_on_step_begin",
    "cb_on_step_end",
    "cb_on_loader_exhausted",
    # validation loop callbacks
    "cb_on_val_begin",
    "cb_on_val_end",
    "cb_on_val_step_begin",
    "cb_on_val_step_end",
    # test loop callbacks
    "cb_on_test_begin",
    "cb_on_test_end",
    "cb_on_test_step_begin",
    "cb_on_test_step_end",
]


class CallbackSignature:
    @abstractmethod
    def cb_on_train_begin(self, trainer: "ModularTaskTrainer") -> None:
        """Called at the beginning of the training loop before the first
        iteration.

        Args:
            trainer: Mutable reference to the trainer.
        """

    @abstractmethod
    def cb_on_train_end(self, trainer: "ModularTaskTrainer") -> None:
        """Called at the end of the training loop after the last iteration,
        validation, and testing are completed.

        Args:
            trainer: Mutable reference to the trainer.
        """

    @abstractmethod
    def cb_on_iteration_begin(
        self,
        trainer: "ModularTaskTrainer",
        iteration: int,
    ) -> None:
        """Called at the beginning of each iteration.

        Args:
            trainer: Mutable reference to the trainer.
            iteration: Current iteration number. For epoch-based training,
                this is the epoch number. For step-based training, this is the
                step number.
        """

    @abstractmethod
    def cb_on_iteration_end(
        self,
        trainer: "ModularTaskTrainer",
        iteration: int,
        metrics: dict,
    ) -> None:
        """Called at the end of each iteration including validation.

        Args:
            trainer: Mutable reference to the trainer.
            iteration: Current iteration number. For epoch-based training,
                this is the epoch number. For step-based training, this is the
                step number.
            metrics: Dictionary of various metrics collected during the
                iteration.
        """

    @abstractmethod
    def cb_on_step_begin(
        self,
        trainer: "ModularTaskTrainer",
        iteration: int,
        batch_idx: int,
    ) -> None:
        """Called at the beginning of step within an iteration.

        Args:
            trainer: Mutable reference to the trainer.
            iteration: Current iteration number. For epoch-based training,
                this is the epoch number. For step-based training, this is the
                step number.
            batch_idx: Current batch index within the iteration. For
                epoch-based training, this is the batch index within the epoch.
                For step-based training, this is the step number modulo the
                evaluation frequency.
        """

    @abstractmethod
    def cb_on_step_end(
        self,
        trainer: "ModularTaskTrainer",
        iteration: int,
        batch_idx: int,
        loss: float,
    ) -> None:
        """Called at the end of step within an iteration.

        Args:
            trainer: Mutable reference to the trainer.
            iteration: Current iteration number. For epoch-based training,
                this is the epoch number. For step-based training, this is the
                step number.
            batch_idx: Current batch index within the iteration. For
                epoch-based training, this is the batch index within the epoch.
                For step-based training, this is the step number modulo the
                evaluation frequency.
            loss: Reduced loss value for the batch.
        """

    @abstractmethod
    def cb_on_loader_exhausted(
        self, trainer: "ModularTaskTrainer", iteration: int
    ) -> None:
        """Called when the training data loader is exhausted.

        Args:
            trainer: Mutable reference to the trainer.
            iteration: Current iteration number. For epoch-based training,
                this is the epoch number. For step-based training, this is the
                step number.
        """

    @abstractmethod
    def cb_on_val_begin(
        self,
        trainer: "ModularTaskTrainer",
        iteration: int,
    ) -> None:
        """Called at the beginning of the validation loop.

        Args:
            trainer: Mutable reference to the trainer.
            iteration: Current iteration number. For epoch-based training,
                this is the epoch number. For step-based training, this is the
                step number.
        """

    @abstractmethod
    def cb_on_val_end(
        self,
        trainer: "ModularTaskTrainer",
        iteration: int,
        val_results: dict,
    ) -> None:
        """Called at the end of the validation loop.

        Args:
            trainer: Mutable reference to the trainer.
            iteration: Current iteration number. For epoch-based training,
                this is the epoch number. For step-based training, this is the
                step number.
            val_results: Dictionary of validation results for the entire
                validation loop of the current iteration.
        """

    @abstractmethod
    def cb_on_val_step_begin(
        self, trainer: "ModularTaskTrainer", batch_idx: int
    ) -> None:
        """Called at the beginning of the validation step.

        Args:
            trainer: Mutable reference to the trainer.
            batch_idx: Current batch index within the validation loop.
        """

    @abstractmethod
    def cb_on_val_step_end(
        self,
        trainer: "ModularTaskTrainer",
        batch_idx: int,
        loss: float,
    ) -> None:
        """Called at the end of the validation step.

        Args:
            trainer: Mutable reference to the trainer.
            batch_idx: Current batch index within the validation loop.
            loss: Reduced loss value for the batch.
        """

    @abstractmethod
    def cb_on_test_begin(self, trainer: "ModularTaskTrainer") -> None:
        """Called at the beginning of the testing loop.

        Args:
            trainer: Mutable reference to the trainer.
        """

    @abstractmethod
    def cb_on_test_end(
        self,
        trainer: "ModularTaskTrainer",
        test_results: dict,
    ) -> None:
        """Called at the end of the testing loop.

        Args:
            trainer: Mutable reference to the trainer.
            test_results: Dictionary of test results for the entire testing
                loop.
        """

    @abstractmethod
    def cb_on_test_step_begin(
        self,
        trainer: "ModularTaskTrainer",
        batch_idx: int,
    ) -> None:
        """Called at the beginning of the testing step.

        Args:
            trainer: Mutable reference to the trainer.
            batch_idx: Current batch index within the testing loop.
        """

    @abstractmethod
    def cb_on_test_step_end(
        self,
        trainer: "ModularTaskTrainer",
        batch_idx: int,
        loss: float,
    ) -> None:
        """

        Args:
            trainer: Mutable reference to the trainer.
            batch_idx: Current batch index within the testing loop.
            loss: Reduced loss value for the batch.
        """


class CallbackManager:
    def __init__(self):
        self.callbacks = {cb: [] for cb in CALLBACK_FUNCTIONS}

    def register(self, obj: Optional[object] = None) -> None:
        if not obj:
            return
        for callback_name in CALLBACK_FUNCTIONS:
            obj_cb_function = getattr(obj, callback_name, None)
            if not obj_cb_function:
                continue
            self.callbacks[callback_name].append(obj_cb_function)

    def register_multiple(self, objs: List[Optional[object]]) -> None:
        for obj in objs:
            self.register(obj)

    def callback(self, position: str, **kwargs) -> None:
        if position not in self.callbacks:
            raise ValueError(f"Callback position {position} not found.")
        for cb in self.callbacks[position]:
            cb(**kwargs)



from typing import Optional

import torch
import torchaudio.transforms as T
import torchvision.transforms.functional as F

from autrainer.core.structs import AbstractDataItem

from .abstract_augmentation import AbstractAugmentation
from .spectrogram_warp_utils import _sparse_image_warp


class GaussianNoise(AbstractAugmentation):
    def __init__(
        self,
        mean: float = 0.0,
        std: float = 1.0,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Add Gaussian noise to the input tensor with mean and standard
        deviation.

        Args:
            mean: The mean of the Gaussian noise. Defaults to 0.0.
            std: The standard deviation of the Gaussian noise. Defaults to 1.0.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(order, p, generator_seed)
        self.mean = mean
        self.std = std
        self._generator = torch.Generator()
        if generator_seed is not None:
            self._generator.manual_seed(generator_seed)

    def offset_generator_seed(self, offset: int) -> None:
        super().offset_generator_seed(offset)
        if self.generator_seed is not None:
            self._generator.manual_seed(self.generator_seed)

    def apply(self, item: AbstractDataItem) -> AbstractDataItem:
        r = torch.randn(item.features.size(), generator=self._generator)
        item.features = item.features + r * self.std + self.mean
        return item


class TimeShift(AbstractAugmentation):
    def __init__(
        self,
        axis: int,
        time_steps: int = 0,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Shift the input tensor along the time axis.

        Args:
            axis: Time axis. If the image is torch Tensor, it is expected
                to have [C, H, W] shape, then H is assumed to be axis 0, and W
                is axis 1.
            time_steps: maximum time steps a tensor will shifted
                forward or backward. Defaults to 0.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        if time_steps < 0:
            raise ValueError(f"Time steps '{time_steps}' must be >= 0.")
        super().__init__(order, p, generator_seed)
        self.axis = axis
        self.time_steps = time_steps
        self._generator = torch.Generator()
        if generator_seed is not None:
            self._generator.manual_seed(generator_seed)

    def offset_generator_seed(self, offset: int) -> None:
        super().offset_generator_seed(offset)
        if self.generator_seed is not None:
            self._generator.manual_seed(self.generator_seed)

    def apply(self, item: AbstractDataItem) -> AbstractDataItem:
        if self.time_steps == 0:
            return item

        t = torch.randint(
            -self.time_steps,
            self.time_steps,
            (1,),
            generator=self._generator,
        ).item()
        if self.axis == 1:
            t1, t2 = item.features[:, :, :-t], item.features[:, :, -t:]
        else:
            t1, t2 = item.features[:, :-t, :], item.features[:, -t:, :]
        item.features = torch.cat((t2, t1), dim=self.axis + 1)
        return item


class TimeMask(AbstractAugmentation):
    def __init__(
        self,
        time_mask: int,
        axis: int,
        replace_with_zero: bool = True,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Mask a random number of time steps.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            time_mask: maximum time steps in a tensor will be masked.
            axis: Time axis. If the image is torch Tensor, it is expected
                to have [C, H, W] shape, then H is assumed to be axis 0, and W
                is axis 1.
            replace_with_zero: Fill the mask either with a tensor mean, or 0's.
                Defaults to True.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        if time_mask < 0:
            raise ValueError(f"Time mask '{time_mask}' must be >= 0.")
        super().__init__(order, p, generator_seed)
        self._deterministic = False
        self.time_mask = time_mask
        self.axis = axis
        self.replace_with_zero = replace_with_zero
        self.masking = T.TimeMasking(time_mask_param=self.time_mask)

    def apply(self, item: AbstractDataItem) -> AbstractDataItem:
        if self.time_mask == 0:
            return item
        if self.axis == 0:
            item.features = torch.rot90(item.features, 3, [1, 2])
            item.features = self.masking(item.features)
            item.features = torch.rot90(item.features, 1, [1, 2])
        else:
            item.features = self.masking(item.features)
        return item


class FrequencyMask(AbstractAugmentation):
    def __init__(
        self,
        freq_mask: int,
        axis: int,
        replace_with_zero: bool = True,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Mask a random number of frequency steps.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            freq_mask: maximum frequency steps in a tensor will be masked.
            axis: Frequency axis. If the image is torch Tensor, it is
                expected to have [C, H, W] shape, then H is assumed to be axis 0,
                and W is axis 1.
            replace_with_zero: Fill the mask either with a tensor mean, or 0's.
                Defaults to True.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        if freq_mask < 0:
            raise ValueError(f"Frequency mask '{freq_mask}' must be >= 0.")
        super().__init__(order, p, generator_seed)
        self._deterministic = False
        self.freq_mask = freq_mask
        self.axis = axis
        self.replace_with_zero = replace_with_zero
        self.masking = T.FrequencyMasking(freq_mask_param=self.freq_mask)

    def apply(self, item: AbstractDataItem) -> AbstractDataItem:
        if self.freq_mask == 0:
            return item
        if self.axis == 1:
            item.features = torch.rot90(item.features, 3, [1, 2])
            item.features = self.masking(item.features)
            item.features = torch.rot90(item.features, 1, [1, 2])
        else:
            item.features = self.masking(item.features)
        return item


class TimeWarp(AbstractAugmentation):
    def __init__(
        self,
        axis: int,
        W: int = 10,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """A random point along the time axis passing through the center of
        the image within the time steps (W, tau - W) is to be warped either to
        the left or right by a distance w chosen from a uniform distribution
        from 0 to the time warp parameter W along that line.

        Args:
            axis: Time axis. If the image is torch Tensor, it is expected
                to have [C, H, W] shape, then H is assumed to be axis 0, and W
                is axis 1.
            W: Bound for squishing/stretching. Defaults to 10.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(order, p, generator_seed)
        self.axis = axis
        self.W = W
        self._generator = torch.Generator()
        if generator_seed is not None:
            self._generator.manual_seed(generator_seed)

    def offset_generator_seed(self, offset: int) -> None:
        super().offset_generator_seed(offset)
        if self.generator_seed is not None:
            self._generator.manual_seed(self.generator_seed)

    def apply(self, item: AbstractDataItem) -> AbstractDataItem:
        device = item.features.device
        if self.axis == 0:
            item.features = torch.rot90(item.features, 3, [1, 2])

        _, num_freq_channels, len_time = F.get_dimensions(item.features)

        # random point along the time axis
        pt = (len_time - 2 * self.W) * torch.rand(
            [1],
            dtype=torch.float,
            generator=self._generator,
        ) + self.W

        # source
        # control points on freq-axis
        src_ctr_pt_freq = torch.arange(0, num_freq_channels // 2)
        src_ctr_pt_time = (
            torch.ones_like(src_ctr_pt_freq) * pt
        )  # control points on time-axis
        src_ctr_pts = torch.stack((src_ctr_pt_freq, src_ctr_pt_time), dim=-1)
        src_ctr_pts = src_ctr_pts.float().to(device)

        # Destination
        w = (
            2
            * self.W
            * torch.rand([1], dtype=torch.float, generator=self._generator)
            - self.W
        )  # distance
        dest_ctr_pt_freq = src_ctr_pt_freq
        dest_ctr_pt_time = src_ctr_pt_time + w
        dest_ctr_pts = torch.stack(
            (dest_ctr_pt_freq, dest_ctr_pt_time),
            dim=-1,
        )
        dest_ctr_pts = dest_ctr_pts.float().to(device)

        src_ctr_pt_locations = torch.unsqueeze(src_ctr_pts, 0)
        dest_ctr_pt_locations = torch.unsqueeze(dest_ctr_pts, 0)

        warped_spectro, _ = _sparse_image_warp(
            item.features,
            src_ctr_pt_locations,
            dest_ctr_pt_locations,
        )

        if self.axis == 0:
            warped_spectro = torch.rot90(warped_spectro, 1, [1, 2])
        item.features = warped_spectro
        return item


class SpecAugment(AbstractAugmentation):
    def __init__(
        self,
        time_mask: int = 10,
        freq_mask: int = 10,
        W: int = 50,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """SpecAugment augmentation.
        A combination of time warp, frequency masking, and time masking.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        For more information, see:
        https://arxiv.org/abs/1904.08779

        This implementation differs from PyTorch, as they apply TimeStrech
        instead of TimeWarp. For more information, see:
        https://pytorch.org/audio/master/tutorials/audio_feature_augmentation_tutorial.html#specaugment

        Args:
            time_mask: maximum time steps in a tensor will be masked.
                Defaults to 10.
            freq_mask: maximum frequency steps in a tensor will be masked.
                Defaults to 10.
            W: Bound for squishing/stretching the time axis. Defaults to 50.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
        """
        super().__init__(order, p, generator_seed)
        self._deterministic = False
        self.time_mask = time_mask
        self.freq_mask = freq_mask
        self.W = W
        self._time_warp = TimeWarp(W=self.W, axis=0)
        self._freq_mask = FrequencyMask(
            freq_mask=self.freq_mask,
            replace_with_zero=True,
            axis=1,
        )
        self._time_mask = TimeMask(
            time_mask=self.time_mask,
            replace_with_zero=True,
            axis=0,
        )

    def apply(self, item: AbstractDataItem) -> AbstractDataItem:
        item = self._time_warp(item)
        item = self._freq_mask(item)
        item = self._time_mask(item)
        return item



from abc import ABC, abstractmethod
import functools
import logging
from typing import Callable, Union
import warnings

import numpy as np
import pandas as pd


def ignore_runtime_warning(func):
    @functools.wraps(func)
    def wrapper_ignore_warning(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            return func(*args, **kwargs)

    return wrapper_ignore_warning


class AbstractMetric(ABC):
    def __init__(
        self,
        name: str,
        fn: Callable,
        fallback: float,
        **fn_kwargs: dict,
    ) -> None:
        """Abstract class for metrics.

        Args:
            name: The name of the metric.
            fn: The function to compute the metric.
            fallback: The fallback value if the metric is NaN.
            **fn_kwargs: Additional keyword arguments to pass to the function.
        """
        self.name = name
        self.fn = fn
        self.fallback = fallback
        self.fn_kwargs = fn_kwargs
        self.logger = logging.getLogger(__name__)

    @ignore_runtime_warning
    def __call__(self, *args, **kwargs) -> float:
        """Compute the metric.

        Args:
            *args: Positional arguments for the function.
            **kwargs: Keyword arguments for the function.

        Returns:
            The score.
        """
        try:
            score = self.fn(*args, **kwargs, **self.fn_kwargs)
        except ValueError as e:
            self.logger.warning(
                f"Error computing {self.name} metric due to: {e}. "
                f"Defaulting to fallback value of {self.fallback}."
            )
            return self.fallback
        if np.isnan(score):
            self.logger.warning(
                f"Error computing {self.name} metric as score is NaN. "
                f"Defaulting to fallback value of {self.fallback}."
            )
            return self.fallback
        return float(score)

    @property
    @abstractmethod
    def starting_metric(self) -> float:
        """The starting metric value.

        Returns:
            The starting metric value.
        """

    @property
    @abstractmethod
    def suffix(self) -> str:
        """The suffix of the metric.

        Returns:
            The suffix of the metric.
        """

    @staticmethod
    @abstractmethod
    def get_best(a: Union[pd.Series, np.ndarray]) -> float:
        """Get the best metric value from a series of scores.

        Args:
            a: Pandas series or numpy array of scores.

        Returns:
            Best metric value.
        """

    @staticmethod
    @abstractmethod
    def get_best_pos(a: Union[pd.Series, np.ndarray]) -> int:
        """Get the position of the best metric value from a series of scores.

        Args:
            a: Pandas series or numpy array of scores.

        Returns:
            Position of the best metric value.
        """

    @staticmethod
    @abstractmethod
    def compare(a: Union[int, float], b: Union[int, float]) -> bool:
        """Compare two scores and return True if the first score is better.

        Args:
            a: First score.
            b: Second score.

        Returns:
            True if the first score is better.
        """

    def unitary(self, y_true: np.ndarray, y_pred: np.ndarray) -> float:
        """Unitary evaluation of metric.

        Metric computed for each individual label
        in the case of multilabel classification
        or regression.

        In the default case, computes the metric
        in the same way as globally.

        Args:
            y_true: ground truth values.
            y_pred: prediction values.

        Returns:
            The unitary score.
        """
        return self.__call__(y_true, y_pred)


class BaseAscendingMetric(AbstractMetric):
    def __init__(
        self,
        name: str,
        fn: Callable,
        fallback: float = None,
        **fn_kwargs: dict,
    ) -> None:
        """Base for ascending metrics with higher values being better.

        Args:
            name: The name of the metric.
            fn: The function to compute the metric.
            fallback: The fallback value if the metric is NaN. If None,
                the fallback value is set to -1e32. Defaults to None.
            **fn_kwargs: Additional keyword arguments to pass to the function.
        """
        super().__init__(name, fn, fallback or -1e32, **fn_kwargs)

    @property
    def starting_metric(self) -> float:
        """Ascending metric starting value.

        Returns:
            -1e32
        """
        return -1e32

    @property
    def suffix(self) -> str:
        """Ascending metric suffix.

        Returns:
            "max"
        """
        return "max"

    @staticmethod
    def get_best(a: Union[pd.Series, np.ndarray]) -> float:
        return float(a.max())

    @staticmethod
    def get_best_pos(a: Union[pd.Series, np.ndarray]) -> int:
        if isinstance(a, pd.Series):
            return int(a.idxmax())
        return int(a.argmax())

    @staticmethod
    def compare(a: Union[int, float], b: Union[int, float]) -> bool:
        return a > b


class BaseDescendingMetric(AbstractMetric):
    def __init__(
        self,
        name: str,
        fn: Callable,
        fallback: float = None,
        **fn_kwargs: dict,
    ) -> None:
        """Base for descending metrics with lower values being better.

        Args:
            name: The name of the metric.
            fn: The function to compute the metric.
            fallback: The fallback value if the metric is NaN. If None,
                the fallback value is set to 1e32. Defaults to None.
            **fn_kwargs: Additional keyword arguments to pass to the function.
        """
        super().__init__(name, fn, fallback or 1e32, **fn_kwargs)

    @property
    def starting_metric(self) -> float:
        """Descending metric starting value.

        Returns:
            1e32
        """
        return 1e32

    @property
    def suffix(self) -> str:
        """Descending metric suffix.

        Returns:
            "min"
        """
        return "min"

    @staticmethod
    def get_best(a: Union[pd.Series, np.ndarray]) -> float:
        return float(a.min())

    @staticmethod
    def get_best_pos(a: Union[pd.Series, np.ndarray]) -> int:
        if isinstance(a, pd.Series):
            return int(a.idxmin())
        return int(a.argmin())

    @staticmethod
    def compare(a: Union[int, float], b: Union[int, float]) -> bool:
        return a < b



from enum import Enum
import logging
from typing import Dict, Optional, Type, TypeVar, Union

import hydra
from omegaconf import DictConfig, OmegaConf


_log = logging.getLogger("autrainer.instantiate")

T = TypeVar("T")


class HydraConvertEnum(Enum):
    """Hydra conversion strategy.

    Options:
        NONE: Pass OmegaConf objects (Hydra default).
        PARTIAL: Use primitive types with Structured Configs.
        OBJECT: Use primitive types and Structured Configs as objects.
        ALL: Use primitive types only.

    For more information, see:
    https://hydra.cc/docs/advanced/instantiate_objects/overview/
    """

    NONE = "none"
    PARTIAL = "partial"
    OBJECT = "object"
    ALL = "all"


def instantiate(
    config: Union[DictConfig, Dict],
    instance_of: Optional[Type[T]] = None,
    convert: Optional[HydraConvertEnum] = None,
    recursive: bool = False,
    **kwargs,
) -> T:
    """Instantiate an object from a configuration Dict or DictConfig.

    The config must contain a `_target_` field that specifies a relative import
    path to the object to instantiate.
    If `_target_` is None, returns None.

    Args:
        config: The configuration to instantiate.
        instance_of: The expected type of the instantiated object.
            Defaults to None.
        convert: The conversion strategy to use, one of HydraConvertEnum.
            Convert is only used if the config does not have a `_convert_`
            attribute. If None, uses HydraConvertEnum.ALL. Defaults to None.
        recursive: Whether to recursively instantiate objects. Recursive is
            only used if the config does not have a `_recursive_` field.
            Defaults to False.
        **kwargs: Additional keyword arguments to pass to the object.

    Raises:
        ValueError: If the config does not have a `_target_` field.
        ValueError: If the instantiated object is not an instance of
            `instance_of` and `instance_of` is provided.

    Returns:
        The instantiated object.
    """
    if isinstance(config, DictConfig):
        config = OmegaConf.to_container(config, resolve=True)
    if not config.get("_target_"):
        raise ValueError(f"Missing _target_ field in config: {config}")
    if config.get("_target_") == "None":
        return None
    if not config.get("_convert_"):
        config["_convert_"] = (
            convert.value if convert else HydraConvertEnum.ALL.value
        )
    if not config.get("_recursive_"):
        config["_recursive_"] = recursive
    config.pop("id", None)

    obj = hydra.utils.instantiate(config, **kwargs)
    if instance_of is not None and not isinstance(obj, instance_of):
        instance_of_name = instance_of.__name__
        config_target = config["_target_"]
        raise ValueError(
            f"Instantiated object '{config_target}' "
            f"is not an instance of '{instance_of_name}'."
        )
    return obj


def instantiate_shorthand(
    config: Union[str, DictConfig, Dict],
    instance_of: Optional[Type[T]] = None,
    convert: Optional[HydraConvertEnum] = None,
    recursive: bool = False,
    **kwargs,
) -> T:
    """Instantiate an object from a shorthand configuration.

    A shorthand config is either a string or a dictionary with a single key.
    If config is a string, it should be a python import path.
    If config is a dictionary, the key should be a python import path and the
    value should be a dictionary of keyword arguments.

    Args:
        config: The config to instantiate.
        instance_of: The expected type of the instantiated object.
            Defaults to None.
        convert: The conversion strategy to use, one of HydraConvertEnum.
            Convert is only used if the config does not have a `_convert_`
            attribute. If None, uses HydraConvertEnum.ALL. Defaults to None.
        recursive: Whether to recursively instantiate objects. Recursive is
            only used if the config does not have a `_recursive_` field.
            Defaults to False.
        **kwargs: Additional keyword arguments to pass to the object.

    Raises:
        ValueError: If the config is empty (None or an empty string/dictionary).

    Returns:
        The instantiated object.
    """

    def _instantiate(c: dict):
        return instantiate(c, instance_of, convert, recursive, **kwargs)

    if config is None or not config:
        raise ValueError(f"Shorthand config '{config}' is empty.")
    if isinstance(config, str):
        return _instantiate({"_target_": config})

    key = next(iter(config.keys()))
    if len(config.keys()) == 1:
        return _instantiate({"_target_": key, **(config[key] or {})})

    _log.warning(
        f"Multiple keys found in shorthand config: {config}. "
        "This is likely due to missing indentation."
    )
    config.pop(key)
    return _instantiate({"_target_": key, **config})



import os


def save_requirements(output_directory: str) -> None:
    reqs_output = os.path.join(output_directory, "requirements.txt")
    with open(reqs_output, "w") as f:
        f.write(os.popen("pip freeze").read())



from typing import Optional

from omegaconf import DictConfig, OmegaConf

import autrainer

from .abstract_script import AbstractScript, MockParser
from .utils import (
    add_hydra_args_to_sys,
    catch_cli_errors,
    check_invalid_config_path_arg,
    run_hydra_cmd,
    running_in_notebook,
)


class TrainScript(AbstractScript):
    def __init__(self) -> None:
        super().__init__(
            "train",
            "Launch a training configuration (Hydra).",
            extended_description=(
                "For more information on Hydra's command line line flags, see:\n"
                "https://hydra.cc/docs/advanced/hydra-command-line-flags/."
            ),
            epilog="Example: autrainer train -cn config.yaml",
            unknown_args=True,
        )

    def main(self, args: dict) -> None:
        @autrainer.main("config")
        def main(cfg: DictConfig) -> float:
            import os

            import hydra

            OmegaConf.set_struct(cfg, False)
            OmegaConf.resolve(cfg)
            output_dir = (
                hydra.core.hydra_config.HydraConfig.get().runtime.output_dir
            )

            # ? Skip if run exists and return best tracking metric
            if os.path.exists(os.path.join(output_dir, "metrics.csv")):
                import autrainer
                from autrainer.core.utils import Bookkeeping
                from autrainer.metrics import AbstractMetric

                dev_metrics = OmegaConf.load(
                    os.path.join(output_dir, "_best", "dev.yaml")
                )
                tracking_metric = autrainer.instantiate_shorthand(
                    config=cfg.dataset.tracking_metric,
                    instance_of=AbstractMetric,
                )
                best_metric = dev_metrics[tracking_metric.name]["all"]

                bookkeeping = Bookkeeping(output_dir)
                bookkeeping.log(f"Skipping: {os.path.basename(output_dir)}")

                return best_metric

            # ? Save cfg to output directory
            cfg_path = os.path.join(output_dir, ".hydra", "config.yaml")
            os.makedirs(os.path.dirname(cfg_path), exist_ok=True)
            OmegaConf.save(cfg, cfg_path)

            from autrainer.training import ModularTaskTrainer

            trainer = ModularTaskTrainer(
                cfg=cfg,
                output_directory=output_dir,
            )
            return trainer.train()

        check_invalid_config_path_arg(self.parser)
        main()


@catch_cli_errors
def train(
    override_kwargs: Optional[dict] = None,
    config_name: str = "config",
    config_path: Optional[str] = None,
) -> None:
    """Launch a training configuration.

    Args:
        override_kwargs: Additional Hydra override arguments to pass to the
            train script.
        config_name: The name of the config (usually the file name without the
            .yaml extension). Defaults to "config".
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
    """
    if running_in_notebook():
        run_hydra_cmd("train", override_kwargs, config_name, config_path)
    else:
        add_hydra_args_to_sys(override_kwargs, config_name, config_path)
        script = TrainScript()
        script.parser = MockParser()
        script.main({})



from abc import ABC, abstractmethod
from functools import cached_property, partial
import os
from typing import Callable, Dict, List, Optional, TypeVar, Union

from omegaconf import DictConfig
import pandas as pd
import torch
from torch.utils.data import DataLoader

import autrainer
from autrainer.core.constants import TrainingConstants
from autrainer.core.structs import DataBatch
from autrainer.core.utils import set_seed
from autrainer.metrics import AbstractMetric
from autrainer.transforms import SmartCompose

from .utils import (
    AbstractFileHandler,
    AbstractTargetTransform,
    DatasetWrapper,
    LabelEncoder,
    MinMaxScaler,
    MultiLabelEncoder,
    MultiTargetMinMaxScaler,
)


T = TypeVar("T")


class AbstractDataset(ABC):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        task: str,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Abstract dataset class.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            task: Task of the dataset in
                :const:`~autrainer.core.constants.TrainingConstants.TASKS`.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        self._assert_task(task)
        self.features_subdir = features_subdir
        if self.features_subdir is None:
            self.features_subdir = self.audio_subdir
        self.path = path
        self.features_path = features_path
        if self.features_path is None:
            self.features_path = self.path
        self._assert_directory(self.features_path, self.features_subdir)
        self.seed = seed
        self.task = task
        self.metrics = [self._init_metric(m) for m in metrics]
        self.tracking_metric = self._init_metric(tracking_metric)
        self.index_column = index_column
        self.target_column = target_column
        self.file_type = file_type
        self.file_handler = self._init_file_handler(file_handler)
        self.train_transform = train_transform or SmartCompose([])
        self.dev_transform = dev_transform or SmartCompose([])
        self.test_transform = test_transform or SmartCompose([])
        self.stratify = stratify or []

        self._generator = torch.Generator().manual_seed(self.seed)
        self._assert_stratify()

        self._train_loader_kwargs = {}
        self._dev_loader_kwargs = {}
        self._test_loader_kwargs = {}

    @property
    def audio_subdir(self) -> str:
        """Subdirectory containing audio data.

        Defaults to `default` for our standard format.
        Should be overridden for datasets
        that do not conform to it.
        """
        return "default"

    @staticmethod
    def _assert_task(task: str) -> None:
        if task not in TrainingConstants().TASKS:
            raise ValueError(f"Task '{task}' not supported.")

    @staticmethod
    def _assert_directory(path: str, features_subdir: str) -> None:
        if not os.path.isdir(path):
            raise ValueError(f"Directory path='{path}' does not exist.")
        if not os.path.isdir(os.path.join(path, features_subdir)):
            raise ValueError(
                f"Directory features_subdir='{features_subdir}' "
                f"does not exist in '{path}'."
            )

    def _assert_stratify(self) -> None:
        if self.task == "ml-classification" and self.stratify:
            raise ValueError(
                "Stratify is not supported for 'ml-classification' tasks."
            )

        for column in self.stratify:
            if (
                column not in self.df_dev.columns
                or column not in self.df_test.columns
            ):
                raise ValueError(
                    f"Stratify column '{column}' is not present in "
                    "the development or test dataframes."
                )

    @property
    @abstractmethod
    def target_transform(self) -> AbstractTargetTransform:
        """Get the transform to apply to the target.

        Returns:
            Target transform.
        """

    @cached_property
    def output_dim(self) -> int:
        """Get the output dimension of the dataset.

        Returns:
            Output dimension.
        """
        return len(self.target_transform)

    @staticmethod
    def _init_metric(metric: Union[str, DictConfig, Dict]) -> AbstractMetric:
        return autrainer.instantiate_shorthand(
            config=metric,
            instance_of=AbstractMetric,
        )

    @staticmethod
    def _init_file_handler(
        file_handler: Union[str, DictConfig, Dict],
    ) -> AbstractFileHandler:
        return autrainer.instantiate_shorthand(
            config=file_handler,
            instance_of=AbstractFileHandler,
        )

    @cached_property
    def df_train(self) -> pd.DataFrame:
        """Dataframe for the training set, loaded from `train.csv` by default.

        Returns:
            Training dataframe.
        """
        return pd.read_csv(os.path.join(self.path, "train.csv"))

    @cached_property
    def df_dev(self) -> pd.DataFrame:
        """Dataframe for the development set, loaded from `dev.csv` by default.

        Returns:
            Development dataframe.
        """
        return pd.read_csv(os.path.join(self.path, "dev.csv"))

    @cached_property
    def df_test(self) -> pd.DataFrame:
        """Dataframe for the test set, loaded from `test.csv` by default.

        Returns:
            Test dataframe.
        """
        return pd.read_csv(os.path.join(self.path, "test.csv"))

    def _init_dataset(
        self,
        df: pd.DataFrame,
        transform: SmartCompose,
    ) -> DatasetWrapper:
        """Initialize a wrapper around torch.utils.data.Dataset.

        Args:
            df: Dataframe to use.
            transform: Transform to apply to the features.

        Returns:
            Initialized dataset.
        """
        return DatasetWrapper(
            path=self.features_path,
            features_subdir=self.features_subdir,
            index_column=self.index_column,
            target_column=self.target_column,
            file_type=self.file_type,
            file_handler=self.file_handler,
            df=df,
            transform=transform,
            target_transform=self.target_transform,
        )

    @cached_property
    def train_dataset(self) -> DatasetWrapper:
        """Get the training dataset.

        Returns:
            Training dataset.
        """
        return self._init_dataset(self.df_train, self.train_transform)

    @cached_property
    def dev_dataset(self) -> DatasetWrapper:
        """Get the development dataset.

        Returns:
            Development dataset.
        """
        return self._init_dataset(self.df_dev, self.dev_transform)

    @cached_property
    def test_dataset(self) -> DatasetWrapper:
        """Get the test dataset.

        Returns:
            Test dataset.
        """
        return self._init_dataset(self.df_test, self.test_transform)

    @property
    def default_collate_fn(self) -> Callable:
        return DataBatch.collate

    def create_train_loader(
        self,
        batch_size: int,
        num_workers: int = 0,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: float = 0,
        prefetch_factor: Optional[int] = None,
        pin_memory_device: str = "",
    ) -> DataLoader:
        self.train_transform.setup(self)
        return DataLoader(
            self.train_dataset,
            batch_size=batch_size,
            shuffle=True,
            generator=self._generator,
            collate_fn=self.train_transform.get_collate_fn(self),
            num_workers=num_workers,
            worker_init_fn=partial(
                self._seed_worker,
                seed=self.seed,
                transform=self.train_transform,
            ),
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            prefetch_factor=prefetch_factor,
            persistent_workers=num_workers > 0,
            pin_memory_device=pin_memory_device,
        )

    def create_dev_loader(
        self,
        batch_size: int,
        num_workers: int = 0,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: float = 0,
        prefetch_factor: Optional[int] = None,
        pin_memory_device: str = "",
    ) -> DataLoader:
        self.dev_transform.setup(self)
        return DataLoader(
            self.dev_dataset,
            batch_size=batch_size,
            shuffle=False,
            generator=self._generator,
            collate_fn=self.dev_transform.get_collate_fn(self),
            num_workers=num_workers,
            worker_init_fn=partial(
                self._seed_worker,
                seed=self.seed,
                transform=self.dev_transform,
            ),
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            prefetch_factor=prefetch_factor,
            persistent_workers=num_workers > 0,
            pin_memory_device=pin_memory_device,
        )

    def create_test_loader(
        self,
        batch_size: int,
        num_workers: int = 0,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: float = 0,
        prefetch_factor: Optional[int] = None,
        pin_memory_device: str = "",
    ) -> DataLoader:
        self.test_transform.setup(self)
        return DataLoader(
            self.test_dataset,
            batch_size=batch_size,
            shuffle=False,
            generator=self._generator,
            collate_fn=self.test_transform.get_collate_fn(self),
            num_workers=num_workers,
            worker_init_fn=partial(
                self._seed_worker,
                seed=self.seed,
                transform=self.test_transform,
            ),
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            prefetch_factor=prefetch_factor,
            persistent_workers=num_workers > 0,
            pin_memory_device=pin_memory_device,
        )

    @staticmethod
    def _seed_worker(
        worker_id: int,
        seed: int,
        transform: SmartCompose,
    ) -> None:
        set_seed(seed + worker_id)
        transform.offset_generator_seed(worker_id)

    @staticmethod
    def download(path: str) -> None:
        """Download the dataset. Can be implemented by subclasses, but is not
        required.

        Args:
            path: Path to download the dataset to.
        """


class BaseClassificationDataset(AbstractDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: str,
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Base classification dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            task="classification",
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @cached_property
    def target_transform(self) -> LabelEncoder:
        return LabelEncoder(
            self.df_train[self.target_column].unique().tolist()
        )

    @staticmethod
    def _assert_dev_split(dev_split: float) -> None:
        if not 0 <= dev_split < 1:
            raise ValueError(f"dev_split '{dev_split}' must be in [0, 1).")

    @staticmethod
    def _assert_choice(choice: T, choices: List[T], name: str) -> None:
        if choice not in choices:
            raise ValueError(f"{name} '{choice}' not in {choices}.")


class BaseMLClassificationDataset(AbstractDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: List[str],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        threshold: float = 0.5,
    ) -> None:
        """Base multi-label classification dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            threshold: Threshold for classification. Defaults to 0.5.
        """
        self._assert_threshold(threshold)
        self.threshold = threshold
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            task="ml-classification",
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )
        self._assert_target_column(allowed_columns=self.df_train.columns)

    @staticmethod
    def _assert_threshold(threshold: float) -> None:
        if not 0 <= threshold <= 1:
            raise ValueError(
                f"Threshold '{threshold}' must be between 0 and 1."
            )

    def _assert_target_column(self, allowed_columns: List[str]) -> None:
        if isinstance(self.target_column, str):
            raise ValueError(
                f"Invalid target column '{self.target_column}'. "
                "Must be a list of strings for multi-label classification."
            )
        if not self.target_column:
            raise ValueError("No target columns provided.")

        if not all(c in allowed_columns for c in self.target_column):
            raise ValueError(
                f"Invalid target columns '{self.target_column}'. "
                f"Must be in {allowed_columns}."
            )

    @cached_property
    def target_transform(self) -> MultiLabelEncoder:
        return MultiLabelEncoder(self.threshold, self.target_column)


class BaseRegressionDataset(AbstractDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: str,
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Base regression dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            task="regression",
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @cached_property
    def target_transform(self) -> MinMaxScaler:
        return MinMaxScaler(
            target=self.target_column,
            minimum=self.df_train[self.target_column].min(),
            maximum=self.df_train[self.target_column].max(),
        )


class BaseMTRegressionDataset(AbstractDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: List[str],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Base multi-target regression dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            task="regression",
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @cached_property
    def target_transform(self) -> MultiTargetMinMaxScaler:
        return MultiTargetMinMaxScaler(
            target=self.target_column,
            minimum=self.df_train[self.target_column].min().to_list(),
            maximum=self.df_train[self.target_column].max().to_list(),
        )



from abc import ABC, abstractmethod
from functools import cached_property
from inspect import signature
from typing import Callable, List, Optional, Union

import audobject
import torch


class AbstractModel(torch.nn.Module, audobject.Object, ABC):
    def __init__(
        self,
        output_dim: int,
        transfer: Optional[Union[bool, str]] = None,
    ) -> None:
        """Abstract model class.

        Args:
            output_dim: Output dimension of the model.
            transfer: Whether to load the model with pretrained weights if
                available. May be a boolean or a string representing a truthy
                or falsy value. Defaults to None.
        """
        super().__init__()
        self.output_dim = output_dim
        self.transfer = transfer
        self.inputs  # precompute and verify inputs
        self.embedding_inputs  # precompute and verify embedding inputs

    @abstractmethod
    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        """Get embeddings from the model.

        Args:
            features: Input tensor.

        Returns:
            Embeddings.
        """

    @cached_property
    def inputs(self) -> List[str]:
        """Get the inputs to the model's forward method.

        Returns:
            Model inputs.
        """
        return self._collect_inputs(self.forward)

    @cached_property
    def embedding_inputs(self) -> List[str]:
        """Get the inputs to the model's embedding method.

        Returns:
            Model inputs.
        """
        return self._collect_inputs(self.embeddings)

    def _collect_inputs(self, fn: Callable) -> List[str]:
        names = [v.name for v in signature(fn).parameters.values()]
        fn_name = fn.__name__
        if names[0] != "features":
            raise NameError(
                (
                    f"Model {type(self).__name__} "
                    "does not have 'features' "
                    f"as the first argument of its '{fn_name}' method. "
                    f"Its arguments are: {names}. "
                    f"Please rewrite the '{fn_name}' method accordingly."
                )
            )
        return names

    @abstractmethod
    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """Get model output.

        Args:
            features: Input tensor.

        Returns:
            Model output.
        """
        raise NotImplementedError  # pragma: no cover



import os
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from autrainer.metrics import AbstractMetric

from .plot_base import PlotBase


class PlotMetrics(PlotBase):
    def __init__(
        self,
        output_directory: str,
        training_type: str,
        figsize: tuple,
        latex: bool,
        filetypes: list,
        pickle: bool,
        context: str,
        palette: str,
        replace_none: bool,
        add_titles: bool,
        add_xlabels: bool,
        add_ylabels: bool,
        rcParams: dict,
        metric_fns: List[AbstractMetric],
    ) -> None:
        """Plot the metrics of one or multiple runs.

        Args:
            output_directory: Output directory to save plots to.
            training_type: Type of training in ["Epoch", "Step"].
            figsize: Figure size in inches.
            latex: Whether to use LaTeX in plots. Requires the `latex` package.
                To install all necessary dependencies, run:
                `pip install autrainer[latex]`.
            filetypes: Filetypes to save plots as.
            pickle: Whether to save additional pickle files of the plots.
            context: Context for seaborn plots.
            palette: Color palette for seaborn plots.
            replace_none: Whether to replace "None" in labels with "~".
            add_titles: Whether to add titles to plots.
            add_xlabels: Whether to add x-labels to plots.
            add_ylabels: Whether to add y-labels to plots.
            rcParams: Additional Matplotlib rcParams to set.
            metric_fns: List of metrics to use for plotting.
        """
        super().__init__(
            output_directory,
            training_type,
            figsize,
            latex,
            filetypes,
            pickle,
            context,
            palette,
            replace_none,
            add_titles,
            add_xlabels,
            add_ylabels,
            rcParams,
        )
        self.metric_fns = metric_fns

    def plot_run(self, metrics: pd.DataFrame, std_scale: float = 0.1) -> None:
        """Plot the metrics of a single run.

        Args:
            metrics: DataFrame containing the metrics.
            std_scale: Scale factor for the standard deviation.
                Defaults to 0.1.
        """
        std_columns = metrics.filter(regex="\\.std$").columns
        has_std = not std_columns.empty

        fig, ax = plt.subplots(1, 1, figsize=self.figsize)
        sns.lineplot(
            data=metrics[["train_loss", "dev_loss"]], ax=ax, dashes=False
        )
        if has_std:
            for col in ["train_loss", "dev_loss"]:
                ax.fill_between(
                    metrics.index,
                    metrics[col] - std_scale * metrics[f"{col}.std"],
                    metrics[col] + std_scale * metrics[f"{col}.std"],
                    alpha=0.2,
                )
        self._add_label(ax, self.training_type, "loss")
        self.save_plot(fig, "loss", path="_plots")

        for key in metrics.columns:
            if ".std" in key or key in ["train_loss", "dev_loss", "iteration"]:
                continue
            fig, ax = plt.subplots(1, 1, figsize=self.figsize)
            sns.lineplot(data=metrics[key], ax=ax, dashes=False)
            if has_std:
                ax.fill_between(
                    metrics.index,
                    metrics[key] - std_scale * metrics[f"{key}.std"],
                    metrics[key] + std_scale * metrics[f"{key}.std"],
                    alpha=0.2,
                )
            self._add_label(ax, self.training_type, key)
            self.save_plot(fig, key, path="_plots")

    def plot_metric(
        self,
        metrics: pd.DataFrame,
        metric: str,
        metrics_std: Optional[pd.DataFrame] = None,
        std_scale: float = 0.1,
        max_runs: Optional[int] = None,
    ) -> None:
        """Plot a single metric of multiple runs.

        Args:
            metrics: DataFrame containing the metrics.
            metric: Metric to plot.
            metrics_std: DataFrame containing the standard deviations.
                Defaults to None.
            std_scale: Scale factor for the standard deviation.
                Defaults to 0.1.
            max_runs: Maximum number of best runs to plot. If None, all runs
                are plotted. Defaults to None.
        """
        fig = plt.figure(figsize=self.figsize)

        if max_runs is None:
            max_runs = len(metrics.columns)
        metrics, metrics_std = self._select_top_runs(
            metric, metrics, metrics_std, max_runs
        )

        sns.lineplot(data=metrics, dashes=False)

        if metrics_std is not None:
            for col in metrics_std.columns:
                plt.fill_between(
                    metrics.index,
                    metrics[col] - std_scale * metrics_std[col],
                    metrics[col] + std_scale * metrics_std[col],
                    alpha=0.2,
                )

        self._add_label(plt.gca(), self.training_type, metric)
        path = os.path.join("plots", "training_plots")
        self.save_plot(fig, metric, path)

    def plot_aggregated_bars(
        self,
        metrics_df: pd.DataFrame,
        metric: str,
        subplots_by: int = 0,
        group_by: int = 1,
        split_subgroups: bool = True,
    ):
        """Plot aggregated bar plots for a metric.

        Generate a bar plots from the metrics_df, which are divided
        by the "subplots_by" column, further grouped according to the
        "group_by" column.
        If "split_subgroups" is set to true, each group is further split into
        subgroups based on what comes after a potential "-" in the "group_by"
        entry.
        Finally the "metric" entries are averaged to create the bars and the
        standard deviation is shown as error bars.

        Args:
            metrics_df: DataFrame containing the metrics.
            metric: Metric to plot.
            subplots_by: Column to group the subplots by.
            group_by: Column to group the data by.
            split_subgroups: Whether to split subgroups.
        """
        label_replacement_models = {
            "None": "scratch",
            "pret": "pretrained",
            "T": "transfer",
        }

        # Group metrics by the specified columns
        plot_metrics = metrics_df.groupby(metrics_df.columns[subplots_by])

        # Prepare data for plotting
        df_list = []
        for subplot, plot_dfs in plot_metrics:
            group_metrics = plot_dfs.groupby(plot_dfs.columns[group_by])
            for group, group_df in group_metrics:
                if split_subgroups:
                    # Use the shared prefix for subgroups
                    group_split = group.split("-")
                    group = group_split[0]
                    if len(group_split) > 1:
                        subgroup = group_split[1]
                    else:
                        subgroup = "None"
                values = group_df[metric].dropna().astype(float).values
                if values.size == 0:
                    continue
                m, s = np.mean(values), np.std(values)
                df_list.append(
                    {
                        "Subplot": subplot,
                        "Group": group,
                        "Subgroup": subgroup,
                        "Mean": m,
                        "Std": s,
                    }
                )
        df = pd.DataFrame(df_list)
        num_subplots = len(df)
        fig, ax = plt.subplots(
            nrows=num_subplots,
            ncols=1,
            figsize=(self.figsize[0], 0.5 * self.figsize[1] * num_subplots),
        )
        for i, (subplot) in enumerate(df["Subplot"]):
            if num_subplots > 1:
                ax_obj = ax[i]
            else:
                ax_obj = ax
            plot_df = df[df["Subplot"] == subplot].reset_index(drop=True)
            bar_plot = sns.barplot(
                data=plot_df,
                x="Group",
                y="Mean",
                hue="Subgroup",
                errorbar=None,
                ax=ax_obj,
            )
            for i, row in plot_df.iterrows():
                bar_plot.errorbar(
                    i,
                    row["Mean"],
                    yerr=row["Std"],
                    fmt="none",
                    c="black",
                    capsize=3,
                )
            legend_labels = []
            for subgroup in df["Subgroup"].unique():
                if subgroup in label_replacement_models.keys():
                    legend_labels.append(label_replacement_models[subgroup])
                else:
                    legend_labels.append(subgroup)
            handles, _ = ax_obj.get_legend_handles_labels()
            ax_obj.legend(
                handles,
                legend_labels,
                bbox_to_anchor=(0.9, 1),
                loc="upper left",
            )
            ax_obj.set_xlabel("")
            ax_obj.set_ylabel(metric)
            ax_obj.set_title(subplot)
        plt.tight_layout()
        path = os.path.join("plots", "bar_plots")
        self.save_plot(fig, metric, path)

    def _select_top_runs(
        self,
        metric: str,
        metrics: pd.DataFrame,
        metrics_std: pd.DataFrame,
        max_runs: int,
    ):
        if "loss" in metric:
            top_values = metrics.min()
            ascending_order = True
        else:
            m = next((fn for fn in self.metric_fns if fn.name == metric), None)
            if m is None:
                raise ValueError(f"No metric found for '{metric}'.")
            if m.suffix == "max":
                top_values = metrics.max()
                ascending_order = False
            else:
                top_values = metrics.min()
                ascending_order = True

        top_runs = (
            top_values.sort_values(ascending=ascending_order)
            .head(max_runs)
            .index
        )

        metrics = metrics[top_runs]
        if metrics_std is not None:
            metrics_std = metrics_std[top_runs]

        return metrics, metrics_std



import audmetric

from .abstract_metric import BaseAscendingMetric


class Accuracy(BaseAscendingMetric):
    def __init__(self):
        """Accuracy metric using `audmetric.accuracy`."""
        super().__init__(name="accuracy", fn=audmetric.accuracy)


class UAR(BaseAscendingMetric):
    def __init__(self):
        """Unweighted average recall metric using
        `audmetric.unweighted_average_recall`.
        """
        super().__init__(name="uar", fn=audmetric.unweighted_average_recall)


class F1(BaseAscendingMetric):
    def __init__(self):
        """F1 metric using `audmetric.unweighted_average_fscore`."""
        super().__init__(name="f1", fn=audmetric.unweighted_average_fscore)



from typing import Dict, List, Union

from .abstract_constants import AbstractConstants


class ExportConstants(AbstractConstants):
    """Singleton for managing the export and logging configurations of
    `autrainer`."""

    _name = "ExportConstants"
    _logging_depth = 2
    _ignore_params = [
        "results_dir",
        "experiment_id",
        "model.dataset",
        "training_type",
        "save_frequency",
        "dataset.metrics",
        "plotting",
        "model.transform",
        "dataset.transform",
        "augmentation.steps",
        "loggers",
        "progress_bar",
        "continue_training",
        "remove_continued_runs",
        "save_train_outputs",
        "save_dev_outputs",
        "save_test_outputs",
    ]
    _artifacts = [
        "model_summary.txt",
        "metrics.csv",
        {"config.yaml": ".hydra"},
    ]

    @property
    def LOGGING_DEPTH(self) -> int:
        """Get the depth of logging for configuration parameters.
        Defaults to :attr:`2`.

        Returns:
            Depth of logging for configuration parameters.
        """
        return self._logging_depth

    @LOGGING_DEPTH.setter
    def LOGGING_DEPTH(self, logging_depth: int) -> None:
        """Set the depth of logging for configuration parameters.

        Args:
            logging_depth: Depth of logging for configuration parameters.

        Raises:
            ValueError: If the logging depth is not an integer or is negative.
        """
        self._assert_type(logging_depth, int, "LOGGING_DEPTH")
        if logging_depth < 0:
            raise ValueError("LOGGING_DEPTH must be non-negative.")
        self._logging_depth = logging_depth

    @property
    def IGNORE_PARAMS(self) -> List[str]:
        """Get the ignored configuration parameters for logging.
        Defaults to :attr:`["results_dir", "experiment_id", "model.dataset",
        "training_type", "save_frequency", "dataset.metrics", "plotting",
        "model.transform", "dataset.transform", "augmentation.steps",
        "loggers", "progress_bar", "continue_training",
        "remove_continued_runs", "save_train_outputs", "save_dev_outputs",
        "save_test_outputs"]`.

        Returns:
            Ignored configuration parameters for logging.
        """
        return self._ignore_params

    @IGNORE_PARAMS.setter
    def IGNORE_PARAMS(self, ignore_params: List[str]) -> None:
        """Set the ignored configuration parameters for logging.

        Args:
            ignore_params: Ignored configuration parameters for logging.

        Raises:
            ValueError: If the ignored parameters are not a list of strings.
        """
        self._assert_type(ignore_params, list, "IGNORE_PARAMS")
        for t in ignore_params:
            self._assert_type(t, str, "IGNORE_PARAMS", "in ignored parameters")
        self._ignore_params = ignore_params

    @property
    def ARTIFACTS(self) -> List[Union[str, Dict[str, str]]]:
        """Get the artifacts to log for runs.
        Defaults to :attr:`["model_summary.txt", "metrics.csv",
        {"config.yaml": ".hydra"}]`.

        Returns:
            Artifacts to log for runs.
        """
        return self._artifacts

    @ARTIFACTS.setter
    def ARTIFACTS(self, artifacts: List[Union[str, Dict[str, str]]]) -> None:
        """Set the artifacts to log for runs.

        Args:
            artifacts: Artifacts to log for runs.

        Raises:
            ValueError: If the artifacts are not a list of strings or
                dictionaries.
        """
        self._assert_type(artifacts, list, "ARTIFACTS")
        for a in artifacts:
            if isinstance(a, dict):
                for k, v in a.items():
                    self._assert_type(k, str, "ARTIFACTS", "in artifacts")
                    self._assert_type(v, str, "ARTIFACTS", "in artifacts")
            else:
                self._assert_type(a, str, "ARTIFACTS", "in artifacts")
        self._artifacts = artifacts



from dataclasses import dataclass
import glob
import os
from typing import Optional

import autrainer

from .abstract_script import AbstractScript, MockParser
from .command_line_error import CommandLineError
from .utils import catch_cli_errors


@dataclass
class InferenceArgs:
    model: str
    input: str
    output: str
    checkpoint: str
    device: str
    extension: str
    recursive: bool
    embeddings: bool
    update_frequency: int
    preprocess_cfg: Optional[str]
    window_length: Optional[float]
    stride_length: Optional[float]
    min_length: Optional[float]
    sample_rate: Optional[int]
    trust_remote: bool = False


class InferenceScript(AbstractScript):
    def __init__(self) -> None:
        e = (
            "Example: autrainer inference hf:autrainer/example "
            "input/ output/ -d cuda:0"
        )
        super().__init__(
            "inference",
            "Perform inference on a trained model.",
            epilog=e,
            dataclass=InferenceArgs,
        )

    def add_arguments(self) -> None:
        self.parser.add_argument(
            "model",
            type=str,
            help=(
                "Local path to model directory or Hugging Face link of the "
                "format: `hf:repo_id[@revision][:subdir]#local_dir`. "
                "Should contain at least one state subdirectory, the "
                "`model.yaml`, `file_handler.yaml`, `target_transform.yaml`, "
                "and `inference_transform.yaml` files."
            ),
        )
        self.parser.add_argument(
            "input",
            type=str,
            help=(
                "Path to input directory. Should contain audio files "
                "of the specified extension."
            ),
        )
        self.parser.add_argument(
            "output",
            type=str,
            help=(
                "Path to output directory. Output includes a YAML file with "
                "predictions and a CSV file with model outputs."
            ),
        )
        self.parser.add_argument(
            "-c",
            "--checkpoint",
            type=str,
            metavar="C",
            required=False,
            default="_best",
            help=(
                "Checkpoint to use for evaluation. "
                "Defaults to '_best' (on validation set)."
            ),
        )
        self.parser.add_argument(
            "-d",
            "--device",
            type=str,
            metavar="D",
            required=False,
            default="cpu",
            help=(
                "CUDA-enabled device to use for processing. "
                "Defaults to 'cpu'."
            ),
        )
        self.parser.add_argument(
            "-e",
            "--extension",
            type=str,
            metavar="E",
            required=False,
            default="wav",
            help=(
                "Type of file to look for in the input directory. "
                "Defaults to 'wav'."
            ),
        )
        self.parser.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            help=(
                "Recursively search for files in the input directory. "
                "Defaults to False."
            ),
        )
        self.parser.add_argument(
            "-emb",
            "--embeddings",
            action="store_true",
            help=(
                "Extract embeddings from the model in addition to predictions."
                "For each file, a .pt file with embeddings will be saved."
                "Defaults to False."
            ),
        )
        self.parser.add_argument(
            "-u",
            "--update-frequency",
            type=int,
            default=1,
            metavar="F",
            required=False,
            help=(
                "Frequency of progress bar updates. "
                "If 0, the progress bar will be disabled. Defaults to 1."
            ),
        )
        self.parser.add_argument(
            "-p",
            "--preprocess-cfg",
            type=str,
            metavar="P",
            required=False,
            default="default",
            help=(
                "Preprocessing configuration to apply to input. Can be a path "
                "to a YAML file or the name of the preprocessing configuration "
                "in the local or autrainer 'conf/preprocessing' directory. "
                "If 'default', the default preprocessing configuration used "
                "during training will be applied. If 'None', no preprocessing "
                "will be applied. Defaults to 'default'."
            ),
        )
        self.parser.add_argument(
            "-w",
            "--window-length",
            type=float,
            metavar="W",
            required=False,
            default=None,
            help=(
                "Window length for sliding window inference in seconds. "
                "If None, the entire input will be processed at once. "
                "Defaults to None."
            ),
        )
        self.parser.add_argument(
            "-s",
            "--stride-length",
            type=float,
            metavar="S",
            required=False,
            default=None,
            help=(
                "Stride length for sliding window inference in seconds. "
                "If None, the entire input will be processed at once. "
                "Defaults to None."
            ),
        )
        self.parser.add_argument(
            "-m",
            "--min-length",
            type=float,
            metavar="M",
            required=False,
            default=None,
            help=(
                "Minimum length of audio file to process in seconds. "
                "Files shorter than the minimum length are padded with zeros. "
                "Sample rate has to be specified for padding. "
                "If None, no minimum length is enforced. Defaults to None."
            ),
        )
        self.parser.add_argument(
            "-sr",
            "--sample-rate",
            type=int,
            metavar="SR",
            required=False,
            default=None,
            help=(
                "Sample rate of audio files in Hz. Has to be specified for "
                "sliding window inference. Defaults to None."
            ),
        )
        self.parser.add_argument(
            "-t",
            "--trust-remote",
            action="store_true",
            help=(
                "Whether to trust and execute custom Python code from Hugging "
                "Face models. Warning: This can be dangerous as it may run "
                "arbitrary code on your machine! Defaults to False."
            ),
        )

    def main(self, args: InferenceArgs) -> None:
        args.model = self._assert_model_exists(args)
        self._assert_valid_input(args)
        self._assert_valid_device(args)
        self._assert_valid_checkpoint(args)
        self._assert_and_set_preprocess_path(args)
        self._inference(args)

    def _assert_model_exists(self, args: InferenceArgs) -> str:
        from autrainer.serving import get_model_path

        try:
            return get_model_path(args.model, args.trust_remote)
        except ValueError as e:
            raise CommandLineError(self.parser, str(e), code=1)

    def _assert_checkpoint_exists(self, args: InferenceArgs) -> None:
        path = os.path.join(args.model, args.checkpoint, "model.pt")
        if os.path.exists(path):
            return
        raise CommandLineError(
            self.parser,
            f"Checkpoint '{args.checkpoint}' does not exist.",
            code=1,
        )

    def _assert_valid_input(self, args: InferenceArgs) -> None:
        if not os.path.exists(args.input):
            raise CommandLineError(
                self.parser,
                f"Input '{args.input}' does not exist.",
                code=1,
            )
        if not os.path.isdir(args.input):
            raise CommandLineError(
                self.parser,
                f"Input '{args.input}' is not a directory.",
                code=1,
            )

        pattern = (
            f"**/*.{args.extension}"
            if args.recursive
            else f"*.{args.extension}"
        )
        if not glob.glob(
            os.path.join(args.input, pattern),
            recursive=True,
        ):
            raise CommandLineError(
                self.parser,
                f"No '{args.extension}' files found in '{args.input}'.",
                code=1,
            )

    def _assert_valid_device(self, args: InferenceArgs) -> None:
        if args.device == "cpu" or args.device.startswith("cuda:"):
            return
        raise CommandLineError(
            self.parser,
            (
                f"Invalid device '{args.device}'. "
                "Please specify 'cpu' or 'cuda:<device_id>'."
            ),
        )

    def _assert_valid_checkpoint(self, args: InferenceArgs) -> None:
        path = os.path.join(args.model, args.checkpoint, "model.pt")
        if os.path.exists(path):
            return
        raise CommandLineError(
            self.parser,
            f"Checkpoint '{args.checkpoint}' does not exist.",
            code=1,
        )

    def _assert_and_set_preprocess_path(self, args: InferenceArgs) -> None:
        if args.preprocess_cfg == "default":
            return

        if args.preprocess_cfg == "None" or args.preprocess_cfg is None:
            args.preprocess_cfg = None
            return

        if not args.preprocess_cfg.endswith(".yaml"):
            args.preprocess_cfg += ".yaml"

        if os.path.exists(args.preprocess_cfg):
            return

        local_path = os.path.join("conf", "preprocessing", args.preprocess_cfg)
        if os.path.exists(local_path):
            args.preprocess_cfg = local_path
            return

        lib_path = os.path.join(
            os.path.dirname(autrainer.__path__[0]),
            "autrainer-configurations",
            "preprocessing",
            args.preprocess_cfg,
        )
        if os.path.exists(lib_path):
            args.preprocess_cfg = lib_path
            return

        raise CommandLineError(
            self.parser,
            (
                f"Preprocessing configuration '{args.preprocess_cfg}' "
                "does not exist."
            ),
            code=1,
        )

    def _inference(self, args: InferenceArgs) -> None:
        from autrainer.serving import Inference

        inference = Inference(
            model_path=args.model,
            checkpoint=args.checkpoint,
            device=args.device,
            preprocess_cfg=args.preprocess_cfg,
            window_length=args.window_length,
            stride_length=args.stride_length,
            min_length=args.min_length,
            sample_rate=args.sample_rate,
        )

        results = inference.predict_directory(
            args.input,
            extension=args.extension,
            recursive=args.recursive,
            update_frequency=args.update_frequency,
        )
        inference.save_prediction_yaml(results, args.output)
        inference.save_prediction_results(results, args.output)

        if not args.embeddings:
            return

        results = inference.embed_directory(
            args.input,
            extension=args.extension,
            recursive=args.recursive,
            update_frequency=args.update_frequency,
        )
        inference.save_embeddings(
            results,
            args.output,
            input_extension=args.extension,
        )


@catch_cli_errors
def inference(
    model: str,
    input: str,
    output: str,
    checkpoint: str = "_best",
    device: str = "cpu",
    extension: str = "wav",
    recursive: bool = False,
    embeddings: bool = False,
    update_frequency: int = 1,
    preprocess_cfg: Optional[str] = "default",
    window_length: Optional[float] = None,
    stride_length: Optional[float] = None,
    min_length: Optional[float] = None,
    sample_rate: Optional[int] = None,
    trust_remote: bool = False,
) -> None:
    """Perform inference on a trained model.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        model: Local path to model directory or Hugging Face link of the
            format: `hf:repo_id[@revision][:subdir]#local_dir`.
            Should contain at least one state subdirectory, the
            `model.yaml`, `file_handler.yaml`, `target_transform.yaml`,
            and `inference_transform.yaml` files.
        input: Path to input directory. Should contain audio files of the
            specified extension.
        output: Path to output directory. Output includes a YAML file with
            predictions and a CSV file with model outputs.
        checkpoint: Checkpoint to use for evaluation. Defaults to '_best'.
        device: CUDA-enabled device to use for processing. Defaults to 'cpu'.
        extension: Type of file to look for in the input directory.
            Defaults to 'wav'.
        recursive: Recursively search for files in the input directory.
            Defaults to False.
        embeddings: Extract embeddings from the model in addition to
            predictions. For each file, a .pt file with embeddings will be
            saved. Defaults to False.
        update_frequency: Frequency of progress bar updates. If 0, the progress
            bar will be disabled. Defaults to 1.
        preprocess_cfg: Preprocessing configuration to apply to input. Can be
            a path to a YAML file or the name of the preprocessing
            configuration in the local or autrainer 'conf/preprocessing'
            directory. If "default", the default preprocessing configuration
            used during training will be applied. If None, no preprocessing
            will be applied. Defaults to "default".
        window_length: Window length for sliding window inference in seconds.
            If None, the entire input will be processed at once.
            Defaults to None.
        stride_length: Stride length for sliding window inference in seconds.
            If None, the entire input will be processed at once.
            Defaults to None.
        min_length: Minimum length of audio file to process in seconds.
            Files shorter than the minimum length are padded with zeros.
            Sample rate has to be specified for padding. If None, no minimum
            length is enforced. Defaults to None.
        sample_rate: Sample rate of audio files in Hz. Has to be specified for
            sliding window inference. Defaults to None.
        trust_remote: Whether to trust and execute custom Python code from
            Hugging Face models. Warning: This can be dangerous as it may run
            arbitrary code on your machine! Defaults to False.

    Raises:
        CommandLineError: If the model, input, or preprocessing configuration
            does not exist, or if the device is invalid.
    """
    script = InferenceScript()
    script.parser = MockParser()
    script.main(
        InferenceArgs(
            model,
            input,
            output,
            checkpoint,
            device,
            extension,
            recursive,
            embeddings,
            update_frequency,
            preprocess_cfg,
            window_length,
            stride_length,
            min_length,
            sample_rate,
            trust_remote,
        )
    )



from .dataset_wrapper import DatasetWrapper
from .file_handlers import (
    AbstractFileHandler,
    AudioFileHandler,
    IdentityFileHandler,
    ImageFileHandler,
    NumpyFileHandler,
)
from .target_transforms import (
    AbstractTargetTransform,
    LabelEncoder,
    MinMaxScaler,
    MultiLabelEncoder,
    MultiTargetMinMaxScaler,
)
from .zip_download_manager import ZipDownloadManager


__all__ = [
    "AbstractFileHandler",
    "AbstractTargetTransform",
    "AudioFileHandler",
    "DatasetWrapper",
    "IdentityFileHandler",
    "ImageFileHandler",
    "LabelEncoder",
    "MinMaxScaler",
    "MultiLabelEncoder",
    "MultiTargetMinMaxScaler",
    "NumpyFileHandler",
    "ZipDownloadManager",
]



from functools import cached_property
import os
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig
import pandas as pd

from autrainer.datasets.abstract_dataset import AbstractDataset
from autrainer.datasets.utils import (
    AbstractTargetTransform,
    LabelEncoder,
    MinMaxScaler,
    MultiTargetMinMaxScaler,
)
from autrainer.transforms import SmartCompose


class MSPPodcast(AbstractDataset):
    def __init__(
        self,
        path: str,
        seed: int,
        metrics: List[str],
        tracking_metric: str,
        target_column: str,
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        index_column: str = "FileName",
        features_subdir: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        categories: List[str] = None,
    ) -> None:
        """MSP-Podcast dataset.

        .. warning::
            There are multiple versions available for this dataset.
            We recommend always using the latest one
            (v1.11 at the time of writing)
            but our code is set up to work with all versions
            (at least up to v1.11).

        .. note::
            Note that after v1.7, the dataset features two test sets.
            We only use ``Test1``, as ``Test2`` was found to be biased
            with respect to gender.
            See https://doi.org/10.21437/Interspeech.2019-1708.

        .. note::
            Unlike other datasets which only support classification
            or regression, MSP-Podcast supports both. This is determined
            by picking the appropriate target column.
            ``EmoClass`` corresponds to categorical emotion classification,
            whereas ``EmoAct``, ``EmoVal``, and ``EmoDom`` to dimensional
            emotion regression for activation (arousal), valence, and dominance,
            respectively.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            index_column: Index column of the dataframe.
                Defaults to `FileName`, as in the original data.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            categories: used to filter out specific emotional categories.
                Useful for training on subset of data/classes, such as the classic
                ["A", "H", "N", "S"] 4-class problem found in literature.
                Defaults to None.
        """
        task = (
            "classification" if target_column == "EmoClass" else "regression"
        )
        self.categories = categories
        super().__init__(
            task=task,
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @property
    def audio_subdir(self) -> str:
        """Subdirectory containing audio data.

        Defaults to `Audios` for MSP-Podcast.
        """
        return "Audios"

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """
        Download the MSP-Podcast dataset.

        As this dataset is not publicly-available, please download it manually
        by contacting Prof. Carlos Busso:
        https://ecs.utdallas.edu/research/researchlabs/msp-lab/MSP-Podcast.html

        This function will not do anything.

        For more information on the data, see:
        https://doi.org/10.1109/TAFFC.2017.2736999
        """
        return None

    @cached_property
    def _df(
        self,
    ) -> pd.DataFrame:
        """Load the dataframes.

        Overrides base class so there is no need
        to create separate train/dev/test files.

        Returns:
            Dataframes for training, development, and testing.
        """
        df = pd.read_csv(
            os.path.join(self.path, "Labels", "labels_consensus.csv")
        )
        if self.categories is not None:
            df = df.loc[df["EmoClass"].isin(self.categories)]
        return df.reset_index(drop=True)

    @cached_property
    def df_train(self):
        return self._df.loc[self._df["Split_Set"] == "Train"].reset_index(
            drop=True
        )

    @cached_property
    def df_dev(self):
        return self._df.loc[
            self._df["Split_Set"] == "Development"
        ].reset_index(drop=True)

    @cached_property
    def df_test(self):
        return self._df.loc[self._df["Split_Set"] == "Test1"].reset_index(
            drop=True
        )

    @cached_property
    def target_transform(self) -> AbstractTargetTransform:
        """Get the target transform.

        Determined automatically based on the type of task.

        Returns:
            Target transform.
        """
        if self.task == "classification":
            return LabelEncoder(
                self.df_train[self.target_column].unique().tolist()
            )
        elif self.task == "regression":
            if isinstance(self.target_column, list):
                return MultiTargetMinMaxScaler(
                    target=self.target_column,
                    minimum=self.df_train[self.target_column].min().to_list(),
                    maximum=self.df_train[self.target_column].max().to_list(),
                )
            else:
                return MinMaxScaler(
                    target=self.target_column,
                    minimum=self.df_train[self.target_column].min(),
                    maximum=self.df_train[self.target_column].max(),
                )
        else:
            raise NotImplementedError(
                f"{self.task} not supported for MSPPodcast"
            )



from typing import TYPE_CHECKING, Dict, List, Optional

import numpy as np
import pandas as pd
import torch

from .utils import assert_nonzero_frequency


if TYPE_CHECKING:  # pragma: no cover
    from autrainer.datasets import AbstractDataset


class CrossEntropyLoss(torch.nn.CrossEntropyLoss):
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Wrapper for `torch.nn.CrossEntropyLoss.forward`.

        Converts the targets to `long` if it is a 1D tensor.

        Args:
            x: Batched model outputs.
            y: Targets.

        Returns:
            Loss.
        """
        if y.ndim == 1:
            y = y.long()
        return super().forward(x, y)


class BalancedCrossEntropyLoss(CrossEntropyLoss):
    def setup(self, data: "AbstractDataset") -> None:
        """Calculate balanced weights for the dataset based on the target
        frequency in the training set.

        Args:
            data: Instance of the dataset.
        """
        frequency = (
            data.df_train[data.target_column]
            .map(data.target_transform)
            .value_counts()
            .sort_index()
            .values
        )

        assert_nonzero_frequency(frequency, len(data.target_transform))
        weight = torch.tensor(1 / frequency, dtype=torch.float32)
        self.weight = weight * len(weight) / weight.sum()


class WeightedCrossEntropyLoss(BalancedCrossEntropyLoss):
    def __init__(self, class_weights: Dict[str, float], **kwargs) -> None:
        """Wrapper for `torch.nn.CrossEntropyLoss` with manual class weights.

        The class weights are automatically normalized to sum up to the number
        of classes.

        Args:
            class_weights: Dictionary with class weights corresponding to the
                target labels and their respective weights.
            **kwargs: Additional keyword arguments passed to
                `torch.nn.CrossEntropyLoss`.
        """
        self.class_weights = class_weights
        super().__init__(**kwargs)

    def setup(self, data: "AbstractDataset") -> None:
        """Calculate the class weights based on the provided dictionary.

        Args:
            data: Instance of the dataset.
        """
        values = []

        for label in data.target_transform.labels:
            if label not in self.class_weights:
                raise ValueError(f"Missing class weight for label '{label}'.")
            values.append(self.class_weights[label])

        assert_nonzero_frequency(np.array(values), len(data.target_transform))
        weight = torch.tensor(values, dtype=torch.float32)
        self.weight = weight * len(weight) / weight.sum()


class BCEWithLogitsLoss(torch.nn.BCEWithLogitsLoss):
    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Wrapper for `torch.nn.BCEWithLogitsLoss.forward`.

        Args:
            x: Batched model outputs.
            y: Targets.

        Returns:
            Loss.
        """
        return super().forward(x, y.float())


class BalancedBCEWithLogitsLoss(BCEWithLogitsLoss):
    weights_buffer: torch.Tensor

    def __init__(
        self,
        weight: Optional[torch.Tensor] = None,
        reduction: str = "mean",
    ):
        """Balanced version of `torch.nn.BCEWithLogitsLoss`.

        `pos_weight` is not supported, as the weights are calculated based on
        the target frequency in the training set.

        Args:
            weight: A manual rescaling weight given to the positive class.
                Defaults to None.
            reduction: Specifies the reduction to apply to the output. Defaults
                to 'mean'.
        """
        super().__init__(weight=weight, reduction=reduction)

    def setup(self, data: "AbstractDataset") -> None:
        """Calculate balanced weights for the dataset based on the target
        frequency in the training set.

        Args:
            data: Instance of the dataset.
        """

        def encode(x: pd.Series) -> List[int]:
            return data.target_transform(x.to_list()).tolist()

        frequency = (
            pd.DataFrame(
                data.df_train[data.target_column]
                .apply(encode, axis=1)
                .to_list(),
                columns=data.target_transform.labels,
            )
            .sum(axis=0)
            .values
        )

        assert_nonzero_frequency(frequency, len(data.target_transform))
        weight = torch.tensor(1 / frequency, dtype=torch.float32)
        weight = weight * len(weight) / weight.sum()
        self.register_buffer("weights_buffer", weight)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """Wrapper for `torch.nn.BCEWithLogitsLoss.forward` with balanced
        weights.

        Args:
            x: Batched model outputs.
            y: Targets.

        Returns:
            Loss.
        """
        return super().forward(x, y) * self.weights_buffer.expand_as(y)


class WeightedBCEWithLogitsLoss(BalancedBCEWithLogitsLoss):
    def __init__(self, class_weights: Dict[str, float], **kwargs) -> None:
        """Wrapper for `torch.nn.BCEWithLogitsLoss` with manual class weights.

        The class weights are automatically normalized to sum up to the number
        of classes.

        Args:
            class_weights: Dictionary with class weights corresponding to the
                target labels and their respective weights.
            **kwargs: Additional keyword arguments passed to
                `torch.nn.BCEWithLogitsLoss`.
        """
        self.class_weights = class_weights
        super().__init__(**kwargs)

    def setup(self, data: "AbstractDataset") -> None:
        """Calculate the class weights based on the provided dictionary.

        Args:
            data: Instance of the dataset.
        """
        values = []

        for label in data.target_transform.labels:
            if label not in self.class_weights:
                raise ValueError(f"Missing class weight for label '{label}'.")
            values.append(self.class_weights[label])
        assert_nonzero_frequency(np.array(values), len(data.target_transform))
        weight = torch.tensor(values, dtype=torch.float32)
        weight = weight * len(weight) / weight.sum()
        self.register_buffer("weights_buffer", weight)



from functools import cached_property
import os
from pathlib import Path
import shutil
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig
import pandas as pd

from autrainer.transforms import SmartCompose

from .abstract_dataset import BaseClassificationDataset
from .utils import ZipDownloadManager


FILES = {
    # TUT Acoustic scenes 2016, Development dataset
    "TUT-acoustic-scenes-2016-development.audio.1.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.1.zip?download=1",
    "TUT-acoustic-scenes-2016-development.audio.2.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.2.zip?download=1",
    "TUT-acoustic-scenes-2016-development.audio.3.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.3.zip?download=1",
    "TUT-acoustic-scenes-2016-development.audio.4.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.4.zip?download=1",
    "TUT-acoustic-scenes-2016-development.audio.5.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.5.zip?download=1",
    "TUT-acoustic-scenes-2016-development.audio.6.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.6.zip?download=1",
    "TUT-acoustic-scenes-2016-development.audio.7.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.7.zip?download=1",
    "TUT-acoustic-scenes-2016-development.audio.8.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.audio.8.zip?download=1",
    "TUT-acoustic-scenes-2016-development.error.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.error.zip?download=1",
    "TUT-acoustic-scenes-2016-development.meta.zip": "https://zenodo.org/records/45739/files/TUT-acoustic-scenes-2016-development.meta.zip?download=1",
    # TUT Acoustic scenes 2016, Evaluation dataset
    "TUT-acoustic-scenes-2016-evaluation.audio.1.zip": "https://zenodo.org/records/165995/files/TUT-acoustic-scenes-2016-evaluation.audio.1.zip?download=1",
    "TUT-acoustic-scenes-2016-evaluation.audio.2.zip": "https://zenodo.org/records/165995/files/TUT-acoustic-scenes-2016-evaluation.audio.2.zip?download=1",
    "TUT-acoustic-scenes-2016-evaluation.audio.3.zip": "https://zenodo.org/records/165995/files/TUT-acoustic-scenes-2016-evaluation.audio.3.zip?download=1",
    "TUT-acoustic-scenes-2016-evaluation.meta.zip": "https://zenodo.org/records/165995/files/TUT-acoustic-scenes-2016-evaluation.meta.zip?download=1",
}


class DCASE2016Task1(BaseClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: Optional[str],
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: str,
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        features_path: Optional[str] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        fold: int = 1,
    ) -> None:
        """TUT Acoustic scenes 2016 Task 1 (DCASE2016Task1) dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
                If `None`, defaults to audio subdirectory,
                which is `default` for the standard format,
                but can be overridden in the dataset specification.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            features_path: Root path to features. Useful
                when features need to be extracted and stored
                in a different directory than the root of the dataset.
                If `None`, will be set to `path`. Defaults to `None`.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            fold: Fold to use in [1, 2, 3, 4]. Defaults to 1.
        """
        self._assert_choice(fold, [1, 2, 3, 4], "fold")
        self.fold = fold
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            features_path=features_path,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @cached_property
    def df_train(self) -> pd.DataFrame:
        return pd.read_csv(
            os.path.join(self.path, f"fold{self.fold}_train.csv")
        )

    @cached_property
    def df_dev(self) -> pd.DataFrame:
        return pd.read_csv(
            os.path.join(self.path, f"fold{self.fold}_evaluate.csv")
        )

    @cached_property
    def df_test(self) -> pd.DataFrame:
        return pd.read_csv(os.path.join(self.path, "test.csv"))

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """Download the TUT Acoustic scenes 2016 Task 1 (DCASE2016Task1)
        dataset.

        For more information on the dataset and dataset split, see:
        https://dcase.community/challenge2016/task-acoustic-scene-classification

        Args:
            path: Path to the directory to download the dataset to.
        """
        out_path = os.path.join(path, "default")
        if os.path.isdir(out_path):
            return
        os.makedirs(out_path, exist_ok=True)

        # download and extract files
        dl_manager = ZipDownloadManager(FILES, path)
        dl_manager.download(
            check_exist=[
                "TUT-acoustic-scenes-2016-development",
                "TUT-acoustic-scenes-2016-evaluation",
            ]
        )
        dl_manager.extract(
            check_exist=[
                "TUT-acoustic-scenes-2016-development",
                "TUT-acoustic-scenes-2016-evaluation",
            ]
        )

        # move audio files to the same directory
        dev_path = os.path.join(path, "TUT-acoustic-scenes-2016-development")
        eval_path = os.path.join(path, "TUT-acoustic-scenes-2016-evaluation")
        for file in os.listdir(os.path.join(dev_path, "audio")):
            shutil.move(os.path.join(dev_path, "audio", file), out_path)
        for file in os.listdir(os.path.join(eval_path, "audio")):
            shutil.move(os.path.join(eval_path, "audio", file), out_path)

        # load folds
        fold_names = [
            f
            for f in os.listdir(os.path.join(dev_path, "evaluation_setup"))
            if f.endswith(".txt") and "_test" not in f
        ]
        folds = [
            pd.read_csv(
                os.path.join(dev_path, "evaluation_setup", f),
                sep="\t",
                header=None,
                names=["filename", "scene_label"],
            )
            for f in fold_names
        ]
        fold_names.append("test.txt")
        folds.append(
            pd.read_csv(
                os.path.join(eval_path, "evaluation_setup", "evaluate.txt"),
                sep="\t",
                header=None,
                names=["filename", "scene_label", "full_filename"],
            )
        )

        # remove erroneous files
        erroneous = pd.read_csv(
            os.path.join(dev_path, "error.txt"),
            sep="\t",
            header=None,
            names=["filename", "begin", "end", "error"],
        )
        folds = [
            fold[~fold["filename"].isin(erroneous["filename"])]
            for fold in folds
        ]

        # remove audio from path
        for fold in folds:
            fold.loc[:, "filename"] = fold["filename"].apply(os.path.basename)

        # rename test files to match the development files
        folds[-1].loc[:, "full_filename"] = folds[-1]["full_filename"].apply(
            os.path.basename
        )
        for short_name, full_name in folds[-1][
            ["filename", "full_filename"]
        ].values:
            if not os.path.isfile(os.path.join(out_path, short_name)):
                continue
            shutil.move(
                os.path.join(out_path, short_name),
                os.path.join(out_path, full_name),
            )
        folds[-1]["filename"] = folds[-1]["full_filename"]
        folds[-1].drop(columns=["full_filename"], inplace=True)

        # save folds
        for filename, fold in zip(fold_names, folds):
            fold.to_csv(
                os.path.join(path, Path(filename).with_suffix(".csv")),
                index=False,
            )

        # remove unnecessary files
        shutil.rmtree(dev_path)
        shutil.rmtree(eval_path)



from typing import Optional

import torch
from transformers import ASTConfig
from transformers import ASTModel as ASTBaseModel

from .abstract_model import AbstractModel


class ASTModel(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        num_hidden_layers: int = 12,
        hidden_size: int = 128,
        dropout: float = 0.5,
        transfer: Optional[str] = None,
    ) -> None:
        """Audio Speech Transformer (AST) model. For more information see:
        https://huggingface.co/docs/transformers/v4.31.0/en/model_doc/audio-spectrogram-transformer

        Args:
            output_dim: Output dimension of the model.
            num_hidden_layers: Number of hidden layers in the transformer.
                Defaults to 12.
            hidden_size: Hidden size of the linear layer. Defaults to 128.
            dropout: Dropout rate. Defaults to 0.5.
            transfer: Name of the pretrained model to load. If None, the default
                AST fine-tuned on AudioSet is used. Defaults to None.
                For more information see:
                https://huggingface.co/MIT/ast-finetuned-audioset-10-10-0.4593
        """
        super().__init__(output_dim, transfer)
        self.num_hidden_layers = num_hidden_layers
        self.hidden_size = hidden_size
        self.dropout = dropout
        if self.transfer:  # pragma: no cover
            self.model = ASTBaseModel.from_pretrained(
                self.transfer,
                num_hidden_layers=num_hidden_layers,
            )
        else:
            config = ASTConfig(
                num_hidden_layers=num_hidden_layers,
            )
            self.model = ASTBaseModel(config)
        layers = [
            torch.nn.Linear(self.model.config.hidden_size, hidden_size),
            torch.nn.ReLU(),
            torch.nn.Dropout(dropout),
            torch.nn.Linear(hidden_size, output_dim),
        ]
        self.out = torch.nn.Sequential(*layers)

    def embeddings(self, features: torch.Tensor) -> torch.Tensor:
        return self.model(features).last_hidden_state.mean(1)

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        x = self.embeddings(features)
        x = self.out(x)
        return x



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

import audobject
import torch


class AbstractTargetTransform(ABC, audobject.Object):
    def __init__(self) -> None:
        """Abstract target transform for handling target or label
        transformations in a dataset.

        Serves as the base for creating custom target transforms that handle
        encoding and decoding targets or labels of a dataset, obtaining
        predictions from a batch of model outputs, and determining the majority
        vote from a list of targets.
        """

    def __call__(self, x: Any) -> Any:
        return self.encode(x)

    @abstractmethod
    def __len__(self) -> int:
        """Get the number of unique targets or labels.

        Returns:
            Number of unique targets or labels.
        """

    @abstractmethod
    def encode(self, x: Any) -> Any:
        """Encode a target or label.

        Args:
            x: Target or label.

        Returns:
            Encoded target or label.
        """

    @abstractmethod
    def decode(self, x: Any) -> Any:
        """Decode a target or label. Serve as the inverse operation of encode.

        Args:
            x: Encoded target or label.

        Returns:
            Decoded target or label.
        """

    @abstractmethod
    def probabilities_training(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs
        during training.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """

    @abstractmethod
    def probabilities_inference(self, x: torch.Tensor) -> torch.Tensor:
        """Get the encoded probabilities from a batch of model outputs.

        Args:
            x: Batch of model outputs.

        Returns:
            Encoded probabilities.
        """

    @abstractmethod
    def predict_inference(self, x: torch.Tensor) -> Union[List[Any], Any]:
        """Get the encoded predictions from a batch of model output
        probabilities.

        Args:
            x: Batch of model output probabilities.

        Returns:
            Encoded predictions.
        """

    @abstractmethod
    def majority_vote(self, x: List[Any]) -> Any:
        """Get the majority vote from a list of decoded targets or labels.

        The majority vote is defined by the subclasses and may be the most
        frequent target, the average of all targets, or any other operation
        that determines a majority vote based on the list of targets or labels.

        Args:
            x: List of decoded targets or labels.

        Returns:
            Decoded majority vote.
        """

    @abstractmethod
    def probabilities_to_dict(self, x: torch.Tensor) -> Dict[str, float]:
        """Convert a tensor of probabilities to a dictionary of targets or
        labels and their probabilities.

        Args:
            x: Tensor of probabilities.

        Returns:
            Dictionary of targets or labels and their probabilities.
        """



from copy import deepcopy
import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple

import hydra
from omegaconf import DictConfig, OmegaConf
import pandas as pd
import torch
from tqdm import tqdm

import autrainer
from autrainer.augmentations import AugmentationManager
from autrainer.core.plotting import PlotMetrics
from autrainer.core.structs import AbstractDataBatch
from autrainer.core.utils import (
    Bookkeeping,
    ThreadManager,
    Timer,
    save_hardware_info,
    save_requirements,
    set_device,
    set_seed,
)
from autrainer.datasets import AbstractDataset
from autrainer.datasets.utils import AbstractFileHandler, AudioFileHandler
from autrainer.loggers import AbstractLogger
from autrainer.models import AbstractModel
from autrainer.models.utils import create_model_inputs
from autrainer.transforms import SmartCompose, TransformManager

from .callback_manager import CallbackManager
from .continue_training import ContinueTraining
from .outputs_tracker import OutputsTracker, init_trackers
from .utils import (
    disaggregated_evaluation,
    format_results,
    load_pretrained_model_state,
    load_pretrained_optim_state,
)


class ModularTaskTrainer:
    def __init__(
        self,
        cfg: DictConfig,
        output_directory: str,
        experiment_id: str = None,
        run_name: str = None,
    ) -> None:
        """Trainer managing the training of a model given a configuration.

        Args:
            cfg: Run configuration.
            output_directory: Output directory for the run.
            experiment_id: Experiment ID for the run. If None, the ID is
                automatically set based on the parent directory of the output
                directory. Defaults to None.
            run_name: Run name for the run. If None, the name is automatically
                set based on the output directory. Defaults to None.
        """
        self._thread_manager = ThreadManager()
        self._cfg = cfg
        self._cfg.criterion = self._cfg.dataset.pop("criterion")

        if isinstance(self.cfg.seed, str):
            training_seed, dataset_seed = map(int, self.cfg.seed.split("-"))
        else:
            training_seed = dataset_seed = self.cfg.seed
        set_seed(training_seed)
        self.DEVICE = set_device(self.cfg.device)
        save_hardware_info(output_directory, device=self.DEVICE)
        self.output_directory = Path(output_directory)
        self.initial_iteration = 1

        # ? Save current requirements.txt
        self._thread_manager.spawn(save_requirements, self.output_directory)

        # ? Load Model and Dataset Transforms and Augmentations
        model_config = self.cfg.model
        dataset_config = self.cfg.dataset

        self._loader_kwargs = {}
        for l in {"train", "dev", "test"}:
            key = f"{l}_loader_kwargs"
            self._loader_kwargs[l] = dataset_config.pop(
                key,
                self.cfg.get(key, {}),
            )

        augmentation_manager = AugmentationManager(self.cfg.augmentation)
        train_aug, dev_aug, test_aug = augmentation_manager.get_augmentations()

        transform_manager = TransformManager(
            model_transform=model_config.pop("transform", None),
            dataset_transform=dataset_config.pop("transform", None),
            train_augmentation=train_aug,
            dev_augmentation=dev_aug,
            test_augmentation=test_aug,
        )

        transforms = transform_manager.get_transforms()
        train_transform, dev_transform, test_transform = transforms

        # ? Load Dataset
        self.data = autrainer.instantiate(
            config=dataset_config,
            instance_of=AbstractDataset,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            seed=dataset_seed,
        )

        # ? Create Bookkeeping
        self.bookkeeping = Bookkeeping(
            output_directory=output_directory,
            file_handler_path=os.path.join(output_directory, "training.log"),
        )

        # ? Misc Training Parameters
        self.disable_progress_bar = not self.cfg.get("progress_bar", False)

        self.criterion = autrainer.instantiate_shorthand(
            config=self.cfg.criterion,
            instance_of=torch.nn.modules.loss._Loss,
            reduction="none",
        )
        if hasattr(self.criterion, "setup"):
            self.criterion.setup(self.data)
        self.criterion.to(self.DEVICE)

        # ? Load Pretrained Model, Optimizer, and Scheduler Checkpoints
        model_checkpoint = model_config.pop("model_checkpoint", None)
        if model_checkpoint and model_config.get("transfer", None):
            model_config.pop("transfer", None)  # skip transfer for checkpoints
        optimizer_checkpoint = model_config.pop("optimizer_checkpoint", None)
        scheduler_checkpoint = model_config.pop("scheduler_checkpoint", None)
        skip_last_layer = model_config.pop("skip_last_layer", True)

        # ? Load Model
        self.output_dim = self.data.output_dim
        self.model = autrainer.instantiate(
            config=model_config,
            instance_of=AbstractModel,
            output_dim=self.output_dim,
        )

        if model_checkpoint:
            state_dict = torch.load(
                model_checkpoint,
                map_location="cpu",
                weights_only=True,
            )
            load_pretrained_model_state(
                self.model,
                state_dict,
                skip_last_layer,
            )

        self._thread_manager.spawn(
            self.bookkeeping.save_model_summary,
            deepcopy(self.model),
            self.data.train_dataset[0].features.unsqueeze(0).shape,
            self.DEVICE,
            "model_summary.txt",
        )

        # ? Load Optimizer
        self.optimizer = autrainer.instantiate(
            config=self.cfg.optimizer,
            instance_of=torch.optim.Optimizer,
            params=self.model.parameters(),
            lr=self.cfg.learning_rate,
        )
        if optimizer_checkpoint:
            state_dict = torch.load(
                optimizer_checkpoint,
                map_location=self.DEVICE,
                weights_only=True,
            )
            load_pretrained_optim_state(
                self.optimizer,
                state_dict,
                skip_last_layer,
            )

        # ? Load Scheduler
        if self.cfg.scheduler.id != "None":
            _scheduler_cfg = self.cfg.scheduler
            self.scheduler_frequency = _scheduler_cfg.pop(
                "step_frequency", "evaluation"
            )
            if self.scheduler_frequency not in ["batch", "evaluation"]:
                raise ValueError(
                    f"Scheduler frequency {self.scheduler_frequency} not supported"
                )
            self.scheduler = autrainer.instantiate(
                config=_scheduler_cfg,
                instance_of=torch.optim.lr_scheduler.LRScheduler,
                optimizer=self.optimizer,
            )
            if self.scheduler is not None and scheduler_checkpoint:
                self.scheduler.load_state_dict(
                    torch.load(
                        scheduler_checkpoint,
                        map_location="cpu",
                        weights_only=True,
                    )
                )
        else:
            self.scheduler = None

        # ? Create Dataloaders
        self.train_loader = self.data.create_train_loader(
            batch_size=self.cfg.batch_size,
            **self._loader_kwargs["train"],
        )
        self.dev_loader = self.data.create_dev_loader(
            batch_size=self.cfg.inference_batch_size or self.cfg.batch_size,
            **self._loader_kwargs["dev"],
        )
        self.test_loader = self.data.create_test_loader(
            batch_size=self.cfg.inference_batch_size or self.cfg.batch_size,
            **self._loader_kwargs["test"],
        )

        # ? Take metrics from dataset and add train/dev loss
        metrics = [m.name for m in self.data.metrics] + [
            "train_loss",
            "dev_loss",
        ]
        self.metrics = pd.DataFrame(columns=metrics)
        self.max_dev_metric = self.data.tracking_metric.starting_metric
        self.best_iteration = 1

        # ? Save initial (and best) Model, Optimizer and Scheduler states
        save_tasks = [
            (self.model, "model.pt", "_initial"),
            (self.optimizer, "optimizer.pt", "_initial"),
            (self.scheduler, "scheduler.pt", "_initial"),
            (self.model, "model.pt", "_best"),
            (self.optimizer, "optimizer.pt", "_best"),
            (self.scheduler, "scheduler.pt", "_best"),
        ]
        for task in save_tasks:
            self._thread_manager.spawn(self.bookkeeping.save_state, *task)

        # ? Load and Save Preprocessing Pipeline if specified
        _preprocess_pipe = SmartCompose([])
        _file_handler = self.data.file_handler
        _features_subdir = cfg.dataset.get("features_subdir", "default")
        if (
            not isinstance(self.data.file_handler, AudioFileHandler)
            and _features_subdir != "default"
        ):
            _preprocess = OmegaConf.to_container(
                hydra.compose(f"preprocessing/{_features_subdir}")
            )["preprocessing"]
            _file_handler = autrainer.instantiate_shorthand(
                config=_preprocess["file_handler"],
                instance_of=AbstractFileHandler,
            )
            _preprocess_pipe = SmartCompose(
                [
                    autrainer.instantiate_shorthand(t)
                    for t in _preprocess["pipeline"]
                ]
            )

        save_tasks = [
            (self.data.target_transform, "target_transform.yaml"),
            (self.model, "model.yaml"),
            (self.data.test_transform, "inference_transform.yaml"),
            (self.data.file_handler, "file_handler.yaml"),
            (_preprocess_pipe, "preprocess_pipeline.yaml"),
            (_file_handler, "preprocess_file_handler.yaml"),
        ]
        for task in save_tasks:
            self._thread_manager.spawn(self.bookkeeping.save_audobject, *task)

        # ? Create Timers
        self.train_timer = Timer(output_directory, "train")
        self.dev_timer = Timer(output_directory, "dev")
        self.test_timer = Timer(output_directory, "test")

        # ? Continue run
        if self.cfg.get("continue_training", False):
            self.continue_training = ContinueTraining(
                run_name=run_name or self.output_directory.name,
                remove_continued_runs=self.cfg.get(
                    "remove_continued_runs", False
                ),
            )
        else:
            self.continue_training = None

        # ? Create Loggers
        self.loggers: List[AbstractLogger] = []
        for logger in self.cfg.get("loggers", []):
            self.loggers.append(
                autrainer.instantiate_shorthand(
                    config=logger,
                    instance_of=AbstractLogger,
                    exp_name=experiment_id
                    or self.output_directory.parent.parent.name,
                    run_name=run_name or self.output_directory.name,
                    metrics=self.data.metrics,
                    tracking_metric=self.data.tracking_metric,
                )
            )

        # ? Create Callbacks and Callback Manager
        callbacks = self.cfg.get("callbacks", [])
        self.callbacks = []
        for callback in callbacks:
            self.callbacks.append(
                autrainer.instantiate_shorthand(
                    config=callback,
                    instance_of=object,
                )
            )

        self.callback_manager = CallbackManager()
        self.callback_manager.register_multiple(
            [
                self.data,
                self.model,
                self.optimizer,
                self.scheduler,
                self.criterion,
                *self.loggers,
                *self.callbacks,
                self.continue_training,  # has to be last as it might overwrite other callbacks
            ]
        )

        # ? Create Plot Metrics
        self.plot_metrics = PlotMetrics(
            self.output_directory,
            self.cfg.training_type,
            **self.cfg.plotting,
            metric_fns=self.data.metrics,
        )

        # ? Create Outputs Tracker
        self.dev_tracker, self.test_tracker = init_trackers(
            exports=[self.cfg.save_dev_outputs, self.cfg.save_test_outputs],
            prefixes=["dev", "test"],
            data=self.data,
            bookkeeping=self.bookkeeping,
        )
        if self.cfg.save_train_outputs:
            self.train_tracker = OutputsTracker(
                export=True,
                prefix="train",
                data=self.data,
                bookkeeping=self.bookkeeping,
            )
        else:
            self.train_tracker = None

    def train(self) -> float:
        """Train the model.

        Raises:
            ValueError: If the training type is not supported.

        Returns:
            The best value of the tracking metric.
        """
        # ? Allow optimizers to have custom step functions
        custom_step = getattr(
            self.optimizer, "custom_step", False
        ) and callable(self.optimizer.custom_step)
        self.train_step_fn = (
            self.optimizer.custom_step if custom_step else self._train_step
        )

        self._thread_manager.join()
        self.callback_manager.callback(
            position="cb_on_train_begin", trainer=self
        )

        if self.cfg.training_type == "epoch":
            self.train_epochs()
        elif self.cfg.training_type == "step":
            self.train_steps()
        else:
            raise ValueError(
                f"Training type {self.cfg.training_type} not supported"
            )

        # ? Score best model on test set
        self.bookkeeping.load_state(self.model, "model.pt", "_best")
        self.bookkeeping.load_state(self.optimizer, "optimizer.pt", "_best")
        self.model.to(self.DEVICE)
        self.model.eval()
        self.bookkeeping.create_folder("_test")
        self.test_timer.start()
        test_results = self.evaluate(
            -1,
            "_test",
            self.test_loader,
            self.data.df_test,
            dev_evaluation=False,
            save_to="test_holistic",
            tracker=self.test_tracker,
        )
        self.test_timer.stop()
        self.callback_manager.callback(
            position="cb_on_test_end",
            trainer=self,
            test_results=test_results,
        )
        self.metrics["iteration"] = self.metrics.index
        self.bookkeeping.save_best_results(
            self.metrics,
            "best_results.yaml",
            self.data.metrics,
            self.data.tracking_metric,
            "_best",
        )
        self.bookkeeping.log(
            format_results(
                self.metrics.loc[self.best_iteration]
                .drop("iteration")
                .to_dict(),
                "Best",
                self.cfg.training_type,
                self.best_iteration,
            )
        )
        self.bookkeeping.log(
            format_results(
                test_results,
                "Test",
                self.cfg.training_type,
            )
        )

        # ? Save Timers
        self.train_timer.save()
        self.dev_timer.save()
        self.test_timer.save()

        # ? Plot Metrics
        self.plot_metrics.plot_run(self.metrics)

        self.bookkeeping.save_results_df(self.metrics, "metrics.csv")
        self.callback_manager.callback(
            position="cb_on_train_end", trainer=self
        )
        return self.metrics.loc[self.best_iteration][
            self.data.tracking_metric.name
        ]

    def train_epochs(self):
        """Train the model for a fixed number of epochs."""
        train_loss = []
        pm = (
            self._loader_kwargs["train"].get("pin_memory", False)
            and self.DEVICE.type == "cuda"
        )
        self.train_timer.start()
        for epoch in range(self.initial_iteration, self.cfg.iterations + 1):
            self.callback_manager.callback(
                position="cb_on_iteration_begin", trainer=self, iteration=epoch
            )
            epoch_folder = f"epoch_{epoch}"
            self.bookkeeping.create_folder(epoch_folder)
            self.model.train()
            self.model.to(self.DEVICE)
            for batch_idx, data in enumerate(
                tqdm(
                    self.train_loader,
                    desc="Train",
                    disable=self.disable_progress_bar,
                )
            ):
                data.to(self.DEVICE, non_blocking=pm)
                self.callback_manager.callback(
                    position="cb_on_step_begin",
                    trainer=self,
                    iteration=epoch,
                    batch_idx=batch_idx,
                )
                loss, output = self.train_step_fn(
                    self.model,
                    data,
                    self.criterion,
                    self.data.target_transform.probabilities_training,
                )
                loss = loss.detach()
                output = output.detach()
                reduced_loss = loss.mean().item()
                if self.scheduler and self.scheduler_frequency == "batch":
                    self.scheduler.step()
                if self.train_tracker:
                    self.train_tracker.update(
                        output, data.target, loss, data.index
                    )
                self.callback_manager.callback(
                    position="cb_on_step_end",
                    trainer=self,
                    iteration=epoch,
                    batch_idx=batch_idx,
                    loss=reduced_loss,
                )
                train_loss.append(reduced_loss)
            if epoch % self.cfg.eval_frequency == 0:
                if self.scheduler and self.scheduler_frequency == "evaluation":
                    self.scheduler.step()
                self.train_timer.stop()
                if self.train_tracker:
                    self.train_tracker.save(epoch_folder)
                train_loss = sum(train_loss) / len(train_loss)
                self.metrics.loc[epoch, "train_loss"] = train_loss
                self.dev_timer.start()
                self.evaluate(
                    epoch,
                    epoch_folder,
                    self.dev_loader,
                    self.data.df_dev,
                    tracker=self.dev_tracker,
                )
                self.dev_timer.stop()
                self.callback_manager.callback(
                    position="cb_on_val_end",
                    trainer=self,
                    iteration=epoch,
                    val_results=self.metrics.loc[epoch].to_dict(),
                )
                self.callback_manager.callback(
                    position="cb_on_iteration_end",
                    trainer=self,
                    iteration=epoch,
                    metrics=self.metrics.loc[epoch].to_dict(),
                )
                if epoch < self.cfg.iterations:
                    train_loss = []
                    self.train_timer.start()
            self.callback_manager.callback(
                position="cb_on_loader_exhausted",
                trainer=self,
                iteration=epoch,
            )

    def train_steps(self):
        """Train the model for a fixed number of steps."""
        pbar = tqdm(
            total=self.cfg.eval_frequency,
            desc="Train",
            disable=self.disable_progress_bar,
        )
        step = self.initial_iteration - 1
        self.callback_manager.callback(
            position="cb_on_iteration_begin", trainer=self, iteration=step
        )
        self.train_loader_iter = iter(self.train_loader)
        train_loss = []
        pm = (
            self._loader_kwargs["train"].get("pin_memory", False)
            and self.DEVICE.type == "cuda"
        )
        self.train_timer.start()
        while step < self.cfg.iterations:
            step += 1
            pbar.update(1)
            self.model.train()
            self.model.to(self.DEVICE)
            try:
                data = next(self.train_loader_iter)
            except StopIteration:
                self.callback_manager.callback(
                    position="cb_on_loader_exhausted",
                    trainer=self,
                    iteration=step,
                )
                self.train_loader_iter = iter(self.train_loader)
                data = next(self.train_loader_iter)
            data.to(self.DEVICE, non_blocking=pm)
            self.callback_manager.callback(
                position="cb_on_step_begin",
                trainer=self,
                iteration=step,
                batch_idx=(step - 1) % self.cfg.eval_frequency,
            )
            loss, output = self.train_step_fn(
                self.model,
                data,
                self.criterion,
                self.data.target_transform.probabilities_training,
            )
            loss = loss.detach()
            output = output.detach()
            reduced_loss = loss.mean().item()
            if self.scheduler and self.scheduler_frequency == "batch":
                self.scheduler.step()
            if self.train_tracker:
                self.train_tracker.update(
                    output, data.target, loss, data.index
                )
            self.callback_manager.callback(
                position="cb_on_step_end",
                trainer=self,
                iteration=step,
                batch_idx=(step - 1) % self.cfg.eval_frequency,
                loss=reduced_loss,
            )
            train_loss.append(reduced_loss)
            if step % self.cfg.eval_frequency == 0:
                if self.scheduler and self.scheduler_frequency == "evaluation":
                    self.scheduler.step()
                self.train_timer.stop()
                step_folder = f"step_{step}"
                self.bookkeeping.create_folder(step_folder)
                if self.train_tracker:
                    self.train_tracker.save(step_folder)
                train_loss = sum(train_loss) / len(train_loss)
                self.metrics.loc[step, "train_loss"] = train_loss
                self.dev_timer.start()
                self.evaluate(
                    step,
                    step_folder,
                    self.dev_loader,
                    self.data.df_dev,
                    tracker=self.dev_tracker,
                )
                self.dev_timer.stop()
                self.callback_manager.callback(
                    position="cb_on_val_end",
                    trainer=self,
                    iteration=step,
                    val_results=self.metrics.loc[step].to_dict(),
                )
                self.callback_manager.callback(
                    position="cb_on_iteration_end",
                    trainer=self,
                    iteration=step,
                    metrics=self.metrics.loc[step].to_dict(),
                )
                if step < self.cfg.iterations:
                    train_loss = []
                    pbar.reset()
                    self.train_timer.start()
                    self.callback_manager.callback(
                        position="cb_on_iteration_begin",
                        trainer=self,
                        iteration=step + 1,
                    )

    def _train_step(
        self,
        model: AbstractModel,
        data: AbstractDataBatch,
        criterion: torch.nn.Module,
        probabilities_fn: Callable,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Train the model for a single step.

        Args:
            model: Model to train.
            data: Data to train on.
            criterion: Criterion to optimize.
            probabilities_fn: Function to convert model output to
                probabilities.

        Returns:
            Tuple containing the non-reduced loss and model outputs.
        """
        self.optimizer.zero_grad()
        output = model(**create_model_inputs(model, data))
        loss = criterion(probabilities_fn(output), data.target)
        loss.mean().backward()
        self.optimizer.step()
        return loss, output

    def evaluate(
        self,
        iteration: int,
        iteration_folder: str,
        loader: torch.utils.data.DataLoader,
        df: pd.DataFrame,
        dev_evaluation: bool = True,
        save_to: str = "dev",
        tracker: Optional[OutputsTracker] = None,
    ) -> Dict[str, float]:
        """Evaluate the model on the dev or test set.

        Args:
            iteration: Current iteration.
            iteration_folder: Folder to save the results to.
            loader: Dataloader to evaluate on.
            df: Groundtruth dataframe.
            dev_evaluation: Whether to evaluate on the dev set.
                Defaults to True.
            save_to: Prefix to save the results to. Defaults to "dev".
            tracker: Tracker to save the outputs. Defaults to None.

        Returns:
            Dictionary containing the evaluation results.
        """
        cb_type = "val" if dev_evaluation else "test"
        kwargs = {"iteration": iteration} if dev_evaluation else {}
        self.callback_manager.callback(
            position=f"cb_on_{cb_type}_begin",
            trainer=self,
            **kwargs,
        )
        self.model.eval()
        self.model.to(self.DEVICE)
        lk = self._loader_kwargs["dev" if dev_evaluation else "test"]
        results = self._evaluate(
            loader=loader,
            tracker=tracker,
            iteration_folder=iteration_folder,
            loader_kwargs=lk,
            cb_type=cb_type,
        )
        if dev_evaluation:
            results["dev_loss"] = results.pop("loss")
            # TODO: it's a bit ugly to filter like this
            for key in list(set(self.metrics.columns) - set(["train_loss"])):
                self.metrics.loc[iteration, key] = results[key]
        else:
            test_results = {"test_loss": results["loss"]}
            # TODO: another ugly filter
            for key in list(
                set(self.metrics.columns) - set(["train_loss", "dev_loss"])
            ):
                test_results[f"test_{key}"] = results[key]

        if dev_evaluation:
            self.bookkeeping.log(
                format_results(
                    self.metrics.loc[iteration].to_dict(),
                    "Dev",
                    self.cfg.training_type,
                    iteration,
                )
            )
        if self.data.stratify or isinstance(self.data.target_column, list):
            logging_results = disaggregated_evaluation(
                targets=tracker.targets,
                predictions=tracker.predictions,
                indices=tracker.indices,
                groundtruth=df,
                metrics=self.data.metrics,
                target_column=self.data.target_column,
                stratify=self.data.stratify,
            )
        else:
            logging_results = {
                k: {"all": v}
                for k, v in results.items()
                if not k.endswith("loss")
            }
        if dev_evaluation:
            logging_results["dev_loss"] = {"all": results["dev_loss"]}
            logging_results["iteration"] = iteration
        else:
            logging_results["loss"] = {"all": results["loss"]}

        self.bookkeeping.save_results_dict(
            logging_results, save_to + ".yaml", iteration_folder
        )

        if not dev_evaluation:
            return test_results

        if self.data.tracking_metric.compare(
            results[self.data.tracking_metric.name], self.max_dev_metric
        ):
            self.max_dev_metric = results[self.data.tracking_metric.name]
            self.best_iteration = iteration
            self.bookkeeping.save_state(self.model, "model.pt", "_best")
            self.bookkeeping.save_state(
                self.optimizer, "optimizer.pt", "_best"
            )
            if self.scheduler:
                self.bookkeeping.save_state(
                    self.scheduler, "scheduler.pt", "_best"
                )

            # ? additionally save all best results
            tracker.save("_best", reset=False)
            self.bookkeeping.save_results_dict(
                logging_results, "dev.yaml", "_best"
            )

        if (
            iteration % self.cfg.save_frequency == 0
            or iteration == self.cfg.iterations
        ):
            self.bookkeeping.save_state(
                self.model, "model.pt", iteration_folder
            )
            self.bookkeeping.save_state(
                self.optimizer, "optimizer.pt", iteration_folder
            )
            if self.scheduler:
                self.bookkeeping.save_state(
                    self.scheduler, "scheduler.pt", iteration_folder
                )
        tracker.reset()

    def _evaluate(
        self,
        loader: torch.utils.data.DataLoader,
        tracker: OutputsTracker,
        iteration_folder: str,
        loader_kwargs: Dict[str, Any],
        cb_type: str = "val",
    ) -> Dict[str, float]:
        """Evaluate the model on the dev or test set.

        Args:
            loader: Dataloader to evaluate on.
            tracker: Tracker to save the outputs.
            iteration_folder: Iteration folder to save the results to.
            cb_type: Callback type. Defaults to "val".

        Returns:
            Dictionary containing the results on the dev or test set.
        """
        pm = (
            loader_kwargs.get("pin_memory", False)
            and self.DEVICE.type == "cuda"
        )
        with torch.no_grad():
            losses = 0
            for batch_idx, data in enumerate(
                tqdm(
                    loader,
                    desc="Evaluate" if cb_type == "val" else "Test",
                    disable=self.disable_progress_bar,
                )
            ):
                self.callback_manager.callback(
                    position=f"cb_on_{cb_type}_step_begin",
                    trainer=self,
                    batch_idx=batch_idx,
                )
                data.to(self.DEVICE, non_blocking=pm)
                output = self.model(**create_model_inputs(self.model, data))
                loss = self.criterion(
                    self.data.target_transform.probabilities_training(output),
                    data.target,
                )
                reduced_loss = loss.mean().item()
                losses += reduced_loss
                tracker.update(output, data.target, loss, data.index)
                self.callback_manager.callback(
                    position=f"cb_on_{cb_type}_step_end",
                    trainer=self,
                    batch_idx=batch_idx,
                    loss=reduced_loss,
                )
            losses /= len(loader)
        tracker.save(iteration_folder, reset=False)
        results = {
            "loss": losses,
        }
        for metric in self.data.metrics:
            results[metric.name] = metric(tracker.targets, tracker.predictions)
        return results

    @property
    def cfg(self) -> DictConfig:
        """Return the configuration of the trainer.

        Returns:
            Copy of the configuration.
        """
        return deepcopy(self._cfg)



import logging
import os
import platform
import threading
from typing import Callable, Optional

from omegaconf import OmegaConf
import psutil
import torch


class ThreadManager:
    """Thread manager singleton for tracking spawned threads."""

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialize()
        return cls._instance

    def _initialize(self) -> None:
        self._threads = []
        self._lock = threading.Lock()

    def spawn(self, target: Callable, *args) -> None:
        """Spawn and start a new thread for the target function.

        Args:
            target: Function to run in the thread.
            args: Arguments to pass to the function.
        """
        with self._lock:
            t = threading.Thread(target=target, args=args or ())
            t.start()
            self._threads.append(t)

    def join(self) -> None:
        """Join all spawned threads."""
        with self._lock:
            threads = self._threads[:]
            self._threads.clear()

        for t in threads:
            t.join()

        with self._lock:
            if self._threads:
                return


def get_gpu_info(device: torch.device) -> Optional[dict]:
    """Get GPU information of the current system.

    Args:
        device: Device to get the GPU information from.

    Returns:
        Dictionary containing GPU information. None if no GPU is available.
    """
    if not torch.cuda.is_available():
        return None

    gpu = torch.cuda.get_device_properties(device)
    return {
        "name": gpu.name,
        "memory_gb": round(gpu.total_memory / (1024**3)),
    }


def get_system_info() -> dict:
    """Get system information of the current system.

    Returns:
        Dictionary containing system information.
    """
    s = {
        "os": platform.system(),
        "platform": platform.platform(),
    }
    if os.environ.get("SLURMD_NODENAME"):
        s.update(
            {
                "node": os.environ.get("SLURMD_NODENAME"),
                "memory_gb": round(
                    int(os.environ.get("SLURM_MEM_PER_NODE", 0)) / 1024
                ),
                "cpu_count": int(os.environ.get("SLURM_CPUS_PER_TASK", 0)),
            }
        )
    else:
        s.update(
            {
                "memory_gb": round(psutil.virtual_memory().total / (1024**3)),
                "cpu_count": psutil.cpu_count(),
            }
        )
    return s


def get_hardware_info(device: torch.device) -> dict:
    """Get hardware information of the current system.

    Args:
        device: Device to get the hardware information from.

    Returns:
        Dictionary containing system and GPU information.
    """
    return {
        "system": get_system_info(),
        "gpu": get_gpu_info(device) if device.type == "cuda" else None,
    }


def save_hardware_info(output_directory: str, device: torch.device) -> None:
    """Save hardware information to a hardware.yaml file.

    Args:
        output_directory: Directory to save the hardware information to.
        device: Device to get the hardware information from.
    """
    os.makedirs(output_directory, exist_ok=True)
    OmegaConf.save(
        get_hardware_info(device),
        os.path.join(output_directory, "hardware.yaml"),
    )


def set_device(device_name: str) -> torch.device:
    """Set the device to use. If the specified CUDA device is unavailable,
    automatically fall back to CPU.

    Args:
        device_name: Name of the device to use.

    Returns:
        Specified device if available, otherwise CPU.
    """
    try:
        device = torch.device(device_name)
        torch.tensor(1).to(device)
    except (RuntimeError, AssertionError):
        device = torch.device("cpu")
        logging.warning(
            f"Device '{device_name}' is not available. Falling back to CPU.",
        )
    return device
